# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Tests for itwinai.utils module."""

from pathlib import Path

import pytest

from itwinai.utils import SignatureInspector, flatten_dict, make_config_paths_absolute, to_uri


def test_flatten_dict():
    """Test flatten dict function."""
    dict1 = dict(a=1, b=dict(b1=2, b2=3))

    flattened = flatten_dict(dict1)
    assert flattened.get("a") == 1
    assert flattened.get("b.b1") == 2
    assert flattened.get("b.b2") == 3
    assert len(flattened) == 3


def test_signature_inspector():
    """Test SignatureInspector class."""

    def f(): ...

    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == 0

    def f(*args): ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, *args): ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, bar=123): ...

    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == 2

    def f(foo, *args, bar=123): ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(*args, **kwargs): ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, /, bar, *arg, **kwargs): ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ("foo", "bar")
    assert inspector.min_params_num == 2
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, /, bar, *, hello, **kwargs): ...

    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ("foo", "bar", "hello")
    assert inspector.min_params_num == 3
    assert inspector.max_params_num == SignatureInspector.INFTY


@pytest.mark.parametrize(
    "args,correct_absolute_args",
    [
        (
            ["--config-path=./relative/path", "--some-other-arg"],
            [f"--config-path={Path('./relative/path').resolve()}", "--some-other-arg"],
        ),
        (
            ["--config-path", "./relative/path", "--some-other-arg"],
            ["--config-path", f"{Path('./relative/path').resolve()}", "--some-other-arg"],
        ),
        (
            ["-cp=./relative/path", "--some-other-arg"],
            [f"-cp={Path('./relative/path').resolve()}", "--some-other-arg"],
        ),
        (
            ["-cp", "./relative/path", "--some-other-arg"],
            ["-cp", f"{Path('./relative/path').resolve()}", "--some-other-arg"],
        ),
    ],
)
def test_make_config_paths_absolute(args, correct_absolute_args):
    """Test that make_config_paths_absolute correctly resolves absolute paths."""
    updated_args = make_config_paths_absolute(args)

    assert updated_args == correct_absolute_args


def test_to_uri():
    relative_path = "relative/path/to/file.txt"
    absolute_path = "/absolute/path/to/file.txt"
    s3_uri = "s3://my-bucket/data/file.txt"
    http_uri = "http://example.com/file.txt"

    assert to_uri(relative_path) == str(Path(relative_path).resolve()), "Should be absolute"
    assert to_uri(absolute_path) == absolute_path, "Should remain unchanged"
    assert to_uri(s3_uri) == s3_uri, "Should remain unchanged"
    assert to_uri(http_uri) == http_uri, "Should remain unchanged"
    assert to_uri(Path(relative_path)) == str(Path(relative_path).resolve()), (
        "Should manage Path"
    )



from typing import Any, Dict, Literal, Optional
import torch
import numpy as np
import pandas as pd
from pathlib import Path
from operator import add

from itwinai.torch.inference import TorchModelLoader, TorchPredictor
from itwinai.loggers import Logger
from itwinai.components import monitor_exec
from itwinai.torch.config import TrainingConfiguration

from src.model import ConvVAE
from torch.utils.data import DataLoader, Dataset, TensorDataset
from src.engine import evaluate
from src.initialization import beta, criterion, n_avg, pixel_wise_criterion

class XTClimPredictor(TorchPredictor):
    def __init__(
            self,
            batch_size: int,
            evaluation: Literal["past", "future"] = 'past',
            seasons: Literal["winter_", "spring_", "summer_", "autumn_"] = 'winter_',
            strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
            model_uri: str | None = None,
            model_class: torch.nn.Module | None = None,
            logger: Logger | None = None,
            checkpoints_location: str = "checkpoints",
            name: str | None = None
    ):

        if model_class is None:
            model_class = ConvVAE

        model_loader = TorchModelLoader(model_uri, model_class) if model_uri else None
        super().__init__(
            model=model_loader,
            config={},
            strategy=strategy,
            logger=logger,
            checkpoints_location=checkpoints_location,
            name=name
        )

        self.evaluation = evaluation
        self.batch_size = batch_size
        self.seasons = seasons
        self.config = TrainingConfiguration(batch_size=batch_size, shuffle_test=False)


    def read_data_and_dataloader(
        self,
        input_dir: Path,
        season: str,
        data_type: str
    ) -> tuple[Dataset, int]:
        """Loads dataset from specified directory and returns the dataloader.

        Args:
            input_dir (Path): The path to the directory containing the dataset.
            season (str): The season identifier used for data filenames.
            data_type (str): The type of dataset to load ('train', 'test', 'proj').

        Returns:
            tuple[Dataset, int]: A tuple with the dataset (Dataset) and its length (int).
        """
        time_path = input_dir / f"dates_{data_type}_{season}data_1memb.csv"
        data_path = input_dir / f"preprocessed_1d_{data_type}_{season}data_1memb.npy"

        time_data = pd.read_csv(time_path)
        data = np.load(data_path)

        n_data = len(data)
        dataset = [
            (
                torch.from_numpy(np.reshape(data[i], (2, 32, 32))),
                time_data['0'][i]
            )
            for i in range(n_data)
        ]

        return dataset, n_data

    def evaluate_and_average(
        self,
        model: torch.nn.Module,
        dataloader: DataLoader,
        dataset: Dataset,
        device: str,
        criterion: torch.nn.Module,
        pixel_wise_criterion: torch.nn.Module,
        n_avg: int
    ) -> tuple[np.ndarray, np.ndarray]:
        """Evaluates model on dataset and averages the losses over `n_avg` iterations.

        Args:
            model (nn.Module): The trained model.
            dataloader (DataLoader): The DataLoader for the dataset.
            dataset (Dataset): The dataset to evaluate the model on.
            device (str): Device (e.g., 'cuda', 'cpu').
            criterion (nn.Module): Loss function.
            pixel_wise_criterion (nn.Module): Additional loss function for pixel-wise evaluation.
            n_avg (int): Number of iterations to average over.

        Returns:
            tuple[np.ndarray, np.ndarray]: Tuple containing averaged losses for the dataset.
        """
        avg_losses = 0
        total_losses = []

        for i in range(n_avg):
            avg_loss, _, losses, _ = evaluate(
                model, dataloader, dataset, device, criterion, pixel_wise_criterion
            )
            avg_losses += avg_loss
            total_losses.append(losses)

        total_losses = np.mean(np.stack(total_losses), axis=0)
        avg_losses = avg_losses / n_avg

        return avg_losses, total_losses

    def predict(self) -> Dict[str, Any]:
        """Predict to use evaluate_and_average instead of direct model inference.

        Returns:
            Dict[str, Any]: A dictionary containing average and total losses.
        """

        avg_losses, total_losses = self.evaluate_and_average(
            model=self.model,
            dataloader=self.inference_dataloader,
            dataset=self.inference_dataset,
            device=self.device,
            criterion=criterion,
            pixel_wise_criterion=pixel_wise_criterion,
            n_avg=n_avg
        )

        predictions = {
            "avg_losses": float(avg_losses),
            "total_losses": total_losses.tolist() if isinstance(total_losses, np.ndarray) else total_losses
        }

        return predictions


    def save_to_csv(self, losses: list | np.ndarray, output_dir: Path, filename: str) -> None:
        """Saves to CSV file at specified location.

        Args:
            losses (list | np.ndarray): Losses to be saved.
            output_dir (Path): Directory to save CSV file.
            filename (str): Name of the CSV file.

        Returns:
            None
        """
        pd.DataFrame(losses).to_csv(output_dir / filename)


    @monitor_exec
    def execute(self):

        # Define the base directory for input and output data
        input_dir = Path("input")
        output_dir = Path("outputs")

        if isinstance(self.model, TorchModelLoader):
            self.model = self.model()
        else:
            raise ValueError("Error: model is not an instance of TorchModelLoader.")

        if self.evaluation == 'past':

            trainset, n_train = self.read_data_and_dataloader(input_dir, self.seasons, "train")
            self.inference_dataset = trainset
            train_predictions = super().execute(inference_dataset=self.inference_dataset, model=self.model)

            testset, n_test = self.read_data_and_dataloader(input_dir, self.seasons, "test")
            self.inference_dataset = testset
            test_predictions = super().execute(inference_dataset=self.inference_dataset, model=self.model)

            train_avg_losses = train_predictions["avg_losses"]
            tot_train_losses = train_predictions["total_losses"]
            test_avg_losses = test_predictions["avg_losses"]
            tot_test_losses = test_predictions["total_losses"]

            # Export to CSV
            self.save_to_csv(tot_train_losses, output_dir, f"train_losses_{self.seasons}1d_allssp.csv")
            self.save_to_csv(tot_test_losses, output_dir, f"test_losses_{self.seasons}1d_allssp.csv")

            # Logging
            print('Train average loss:', train_avg_losses)
            print('Test average loss:', test_avg_losses)
            self.log(train_avg_losses, 'Average train loss', kind='metric')
            self.log(test_avg_losses, 'Average test loss', kind='metric')

        else:
            SCENARIO_585 = '585'
            SCENARIO_245 = '245'

            for scenario in [SCENARIO_585, SCENARIO_245]:

                # projection set and data loader
                projset, n_proj = read_data_and_dataloader(input_dir, self.seasons, "proj{scenario}")

                self.config.batch_size = 1

                self.inference_dataset = projset
                proj_predictions = super().execute(inference_dataset=self.inference_dataset, model=self.model)

                proj_avg_losses = proj_predictions["avg_losses"]
                tot_proj_losses = proj_predictions["total_losses"]

                # Export to CSV
                self.save_to_csv(
                    tot_proj_losses,
                    output_dir,
                    f"proj{scenario}_losses_{self.seasons}1d_allssp.csv"
                )
                print(f'SSP{scenario} Projection average loss: {proj_avg_losses} for {self.seasons[:-1]}')

                # Logging
                self.log(proj_avg_losses, 'Average projection loss', kind='metric')



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Show how to use DDP, Horovod and DeepSpeed strategies interchangeably
with a large neural network trained on Imagenet dataset, showing how
to use checkpoints.
"""

import os
import sys
from pathlib import Path
from timeit import default_timer as timer

import horovod.torch as hvd
import torch
import torchvision
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from utils import get_parser, imagenet_dataset, train_epoch

from itwinai.loggers import EpochTimeTracker
from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    TorchDDPStrategy,
)
from itwinai.torch.reproducibility import seed_worker, set_seed


def main():
    # Parse CLI args
    parser = get_parser()
    args = parser.parse_args()

    # Instantiate Strategy
    if args.strategy == "ddp":
        if not torch.cuda.is_available() or not torch.cuda.device_count() > 1:
            raise RuntimeError("Resources unavailable")

        strategy = TorchDDPStrategy(backend=args.backend)
        distribute_kwargs = {}
    elif args.strategy == "horovod":
        strategy = HorovodStrategy()
        distribute_kwargs = dict(
            compression=(
                hvd.Compression.fp16 if args.fp16_allreduce else hvd.Compression.none
            ),
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor,
        )
    elif args.strategy == "deepspeed":
        strategy = DeepSpeedStrategy(backend=args.backend)
        distribute_kwargs = dict(
            config_params=dict(train_micro_batch_size_per_gpu=args.batch_size)
        )
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented."
        )
    strategy.init()

    # Check resource availability
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = use_cuda and torch.cuda.device_count() > 0
    epoch_time_save_dir = Path(args.epoch_time_directory)

    # Dataset
    train_dataset = imagenet_dataset(args.data_dir, subset_size=args.subset_size)

    # Set random seed for reproducibility
    torch_prng = set_seed(args.rnd_seed, deterministic_cudnn=False)

    # Get job rank info - rank==0 master gpu
    if is_distributed:
        # local world size - per node
        global_world_size = strategy.global_world_size()
        global_rank = strategy.global_rank()
    else:
        # Use a single worker (either on GPU or CPU)
        global_world_size = 1
        global_rank = 0

    # Encapsulate the model on the GPU assigned to the current process
    device = torch.device(strategy.device() if use_cuda else "cpu")

    if is_distributed:
        # Distributed sampler restricts data loading to a subset of the dataset
        # exclusive to the current process.
        train_sampler = DistributedSampler(
            train_dataset,
            num_replicas=global_world_size,
            rank=global_rank,
            shuffle=(args.shuff and args.rnd_seed is None),
        )

        train_loader = DataLoader(
            train_dataset,
            batch_size=args.batch_size,
            sampler=train_sampler,
            num_workers=args.nworker,
            pin_memory=True,
            persistent_workers=(args.nworker > 1),
            prefetch_factor=args.prefetch,
            generator=torch_prng,
            worker_init_fn=seed_worker,
        )
    else:
        train_loader = DataLoader(
            train_dataset,
            batch_size=args.batch_size,
            generator=torch_prng,
            worker_init_fn=seed_worker,
        )

    # Create CNN model: resnet 50, resnet101, resnet152
    model = torchvision.models.resnet152()
    model.to(device)

    # Optimizer
    optimizer = torch.optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    if is_distributed:
        model, optimizer, _ = strategy.distributed(
            model, optimizer, lr_scheduler=None, **distribute_kwargs
        )

    if strategy.is_main_worker:
        num_nodes = os.environ.get("SLURM_NNODES", 1)
        strategy_name = f"{args.strategy}-it"
        save_path = epoch_time_save_dir / f"epochtime_{strategy_name}_{num_nodes}.csv"
        epoch_time_logger = EpochTimeTracker(
            strategy_name=strategy_name,
            save_path=save_path,
            num_nodes=int(num_nodes),
        )

    start_time = timer()
    for epoch_idx in range(1, args.epochs + 1):
        epoch_start_time = timer()
        if is_distributed:
            train_sampler.set_epoch(epoch_idx)

        # Training
        train_epoch(
            model=model,
            device=device,
            train_loader=train_loader,
            optimizer=optimizer,
        )

        if strategy.is_main_worker:
            epoch_elapsed_time = timer() - epoch_start_time
            epoch_time_logger.add_epoch_time(epoch_idx, epoch_elapsed_time)
            print(f"[{epoch_idx}/{args.epochs}] - time: {epoch_elapsed_time:.2f}s")

    if global_rank == 0:
        total_time = timer() - start_time
        print(f"Training finished - took {total_time:.2f}s")
        epoch_time_logger.save()

    # Clean-up
    if is_distributed:
        strategy.clean_up()


if __name__ == "__main__":
    main()
    sys.exit()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import logging
import os
import shutil
import tempfile
import time
from pathlib import Path
from typing import Generator

import pytest
import torch
import torch.nn.functional as F
from torch import nn
from torch.utils.data import Subset

from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
)


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=0)


MNIST_PATH = "mnist_dataset"


@pytest.fixture(scope="package")
def ddp_strategy() -> Generator[TorchDistributedStrategy, None, None]:
    """Instantiate Torch's DistributedDataParallel strategy."""
    strategy = TorchDDPStrategy(backend="nccl" if torch.cuda.is_available() else "gloo")
    strategy.init()
    yield strategy
    strategy.clean_up()


@pytest.fixture(scope="package")
def deepspeed_strategy() -> Generator[DeepSpeedStrategy, None, None]:
    """Instantiate DeepSpeed strategy."""
    strategy = DeepSpeedStrategy(backend="nccl" if torch.cuda.is_available() else "gloo")
    strategy.init()
    yield strategy
    strategy.clean_up()


@pytest.fixture(scope="package")
def horovod_strategy() -> Generator[HorovodStrategy, None, None]:
    """Instantiate Horovod strategy."""
    strategy = HorovodStrategy()
    strategy.init()
    yield strategy
    strategy.clean_up()


@pytest.fixture(scope="function")
def shared_tmp_path():
    """Return the Path to a shared filesystem that all nodes can access.
    This is good for Ray-based Trainer because it gets broadcasted to all workers by Ray.
    /tmp location is usually a local filesystem. Uses as a prefix SHARED_FS_PATH
    env variable, but if that's not set it falls back to /tmp.
    """
    if not os.environ.get("SHARED_FS_PATH"):
        logging.warning(
            "SHARED_FS_PATH env var not set! Falling back to /tmp, but this could cause "
            "problems as this fixture should return the path to a location reachable by all "
            "nodes (/tmp usually isn't)."
        )
    else:
        Path(os.environ.get("SHARED_FS_PATH")).mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=os.environ.get("SHARED_FS_PATH", "/tmp")) as tmp_path:
        yield Path(tmp_path)


@pytest.fixture(scope="function")
def named_temp_dir(request):
    """Create a temporay directory for a test case with the name of the test.
    This is useful when you want a tempdir which has a stable path across multiple workers, but
    you are not relying on Ray Trainer to broadcast a single path to all workers.
    """
    user_id = os.getuid()  # Get Linux user ID to avoid collisions among users
    test_name = request.node.name  # Get the name of the calling test

    base_path = Path(os.environ.get("SHARED_FS_PATH", "/tmp"))

    temp_dir = base_path / str(user_id) / test_name
    temp_dir.mkdir(parents=True, exist_ok=True)
    yield temp_dir

    # Cleanup after the test, but wait a bit for all the workers to have accessed the files
    # before deleting them
    time.sleep(5)
    shutil.rmtree(temp_dir, ignore_errors=True)


@pytest.fixture(scope="module")
def mnist_datasets():
    """Parse MNIST datasets."""

    import logging
    import os

    from torchvision import datasets, transforms

    if not os.environ.get("MNIST_PATH"):
        logging.warning("MNIST dataset not found locally. I have to download it!")

    dataset_path = os.environ.get("MNIST_PATH", MNIST_PATH)
    train_set = datasets.MNIST(
        dataset_path,
        train=True,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )
    val_set = datasets.MNIST(
        dataset_path,
        train=False,
        download=False,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )

    # Downsample datasets
    train_set = Subset(train_set, range(500))
    val_set = Subset(val_set, range(500))
    return train_set, val_set


@pytest.fixture(scope="function")
def mnist_net():
    return Net()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


import numpy as np
import torch
from torch.nn import CrossEntropyLoss
from torch.optim.adam import Adam
from torchvision.models import resnet18

from itwinai.torch.distributed import RayDeepSpeedStrategy
from itwinai.torch.trainer import TorchTrainer


class FashionMNISTTrainer(TorchTrainer):
    def create_model_loss_optimizer(self):
        model = resnet18(num_classes=10)
        model.conv1 = torch.nn.Conv2d(
            1, 64, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3), bias=False
        )
        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, RayDeepSpeedStrategy):
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}
        optimizer = Adam(model.parameters(), lr=self.config.optim_lr)
        self.model, self.optimizer, _ = self.strategy.distributed(
            model, optimizer, **distribute_kwargs
        )
        self.loss = CrossEntropyLoss()

    def train(self) -> None:
        device = self.strategy.device()

        for self.current_epoch in range(self.epochs):
            self.set_epoch()

            train_losses = []
            val_losses = []

            # Training epoch
            self.model.train()
            for images, labels in self.train_dataloader:
                images, labels = images.to(device), labels.to(device)
                outputs = self.model(images)
                train_loss = self.loss(outputs, labels)
                self.optimizer.zero_grad()
                train_loss.backward()
                self.optimizer.step()
                train_losses.append(train_loss.detach().cpu().numpy())

            # Validation epoch
            self.model.eval()
            for images, labels in self.validation_dataloader:
                images, labels = images.to(device), labels.to(device)
                with torch.no_grad():
                    outputs = self.model(images)
                    val_loss = self.loss(outputs, labels)
                val_losses.append(val_loss.detach().cpu().numpy())

            # Log metrics with itwinai loggers
            self.log(
                np.mean(train_losses), "train_loss", kind="metric", step=self.current_epoch
            )
            self.log(np.mean(val_losses), "val_loss", kind="metric", step=self.current_epoch)

            # Report metrics and checkpoint to Ray head
            checkpoint = {
                "epoch": self.current_epoch,
                "loss": train_loss,
                "val_loss": val_loss,
            }
            metrics = {"loss": val_loss.item()}
            self.ray_report(metrics=metrics, checkpoint_data=checkpoint)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import sys
from unittest.mock import patch

import pytest

from itwinai.cli import exec_pipeline
from itwinai.pipeline import Pipeline
from itwinai.tests import (
    FakeGetterExec,
    FakeSaverExec,
    FakeSplitterExec,
    FakeTrainerExec,
)
from itwinai.tests.dummy_components import FakePreproc, FakeTrainer


def test_slice_into_sub_pipelines():
    """Test slicing the pipeline to obtain sub-pipelines as Pipeline objects."""
    p = Pipeline(["step1", "step2", "step3"])
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1[0] == "step1"
    assert len(sub_pipe2) == 2

    p = Pipeline(dict(step1="step1", step2="step2", step3="step3"))
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1["step1"] == "step1"
    assert len(sub_pipe2) == 2


@pytest.mark.parametrize(
    "pipeline",
    [
        Pipeline(
            steps=[
                FakePreproc(max_items=33, name="my-preproc"),
                FakeTrainer(lr=0.001, batch_size=64, name="my-trainer"),
            ]
        ),
        Pipeline(
            steps={
                "preproc": FakePreproc(max_items=33, name="my-preproc"),
                "trainer": FakeTrainer(lr=0.001, batch_size=64, name="my-trainer"),
            }
        ),
    ],
)
def test_serialization_to_yaml(pipeline, tmp_path, monkeypatch):
    """Test serialization of pipeline to yaml."""
    yaml_path = tmp_path / "pipe.yaml"
    pipeline.to_yaml(file_path=yaml_path, nested_key="my_pipeline")

    with pytest.raises(SystemExit):
        with (
            patch(
                "itwinai.tests.dummy_components.FakePreproc.__init__", return_value=None
            ) as mock_preproc_init,
            patch(
                "itwinai.tests.dummy_components.FakeTrainer.__init__", return_value=None
            ) as mock_trainer_init,
        ):
            args = [
                "exec_pipeline",
                f"--config-path={tmp_path}",
                "pipe_key=my_pipeline",
            ]
            monkeypatch.setattr(sys, "argv", args)

            exec_pipeline()

            mock_preproc_init.assert_called_once_with(max_items=33, name="my-preproc")
            mock_trainer_init.assert_called_once_with(
                lr=0.001, batch_size=64, name="my-trainer"
            )


def test_arguments_mismatch():
    """Test mismatch of arguments passed among components in a pipeline."""
    pipeline = Pipeline(
        [
            FakeGetterExec(data_uri="http://..."),
            FakeSplitterExec(train_prop=0.7),
            FakeTrainerExec(lr=1e-3, batch_size=32),
            # Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
            FakeSaverExec(save_path="my_model.pth"),
        ]
    )
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too many input arguments" in str(exc_info.value)

    pipeline = Pipeline(
        [
            FakeGetterExec(data_uri="http://..."),
            FakeTrainerExec(lr=1e-3, batch_size=32),
        ]
    )
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too few input arguments" in str(exc_info.value)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Custom types definition."""

from typing import Callable

import torch

#: Torch data batch sampled by a ``DataLoader``.
Batch = torch.Tensor

#: Torch metric function provided by ``torchmetrics`` library.
Metric = Callable


class UninitializedStrategyError(Exception):
    """Error raised when a strategy has not been initialized."""


class DistributedStrategyError(Exception):
    """Error raised when a strategy has already been initialized."""



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from itertools import cycle
from typing import Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch
from matplotlib.ticker import NullLocator, ScalarFormatter

# Doing this because otherwise I get an error about X11 Forwarding which I believe
# is due to the server trying to pass the image to the client computer
matplotlib.use("Agg")

marker_cycle = cycle("ov^s*dXpD.+12348")


def calculate_plot_dimensions(num_datapoints: int) -> Tuple[int, int]:
    """Calculates the height and width of a plot, given a number of datapoints.

    Returns:
        int: The calculated height
        int: The calculated width
    """

    # Note: These calculations are somewhat arbitrary, so this is open to suggestions
    width = max(int(2 * num_datapoints), 8)
    height = min(int(width * 0.8), 6)
    return height, width


def absolute_avg_epoch_time_plot(
    avg_epoch_time_df: pd.DataFrame, gpus_per_node: int = 4
) -> Tuple[Figure, Axes]:
    """Generates a log-log plot of average epoch training times against the number of GPUs
    for distributed training strategies.

    Args:
        avg_epoch_time_df (pd.DataFrame): A DataFrame containing the following columns:
            - "nodes": Number of nodes used in the training process.
            - "avg_epoch_time": Average time (in seconds) taken for an epoch.
            - "name": Name of the distributed training strategy.
        gpus_per_node (int): Number of GPUs per node. Used to calculate the total number
            of GPUs for each training configuration. Defaults to 4.

    Returns:
        Tuple[Figure, Axes]: A tuple containing the matplotlib `Figure` and `Axes` objects
        of the generated plot.

    Raises:
        ValueError: If `avg_epoch_time_df` is missing required columns.
    """
    sns.set_theme()
    fig, ax = plt.subplots()

    unique_nodes = list(avg_epoch_time_df["nodes"].unique())
    unique_names = avg_epoch_time_df["name"].unique()
    num_gpus = [num_node * gpus_per_node for num_node in unique_nodes]
    for name in unique_names:
        data = avg_epoch_time_df[avg_epoch_time_df["name"] == name]

        marker = next(marker_cycle)
        ax.plot(
            data["nodes"] * gpus_per_node,
            data["avg_epoch_time"],
            marker=marker,
            label=name,
            linestyle="-",
            markersize=6,
        )

    # The log scale must be set before changing the ticks
    ax.set_yscale("log")
    ax.set_xscale("log")

    # Calculating upper and lower bounds to get the appropriate top and bottom ticks
    upper_bound = 10 ** int(np.ceil(np.log10(avg_epoch_time_df["avg_epoch_time"].max())))
    lower_bound = 10 ** int(np.floor(np.log10(avg_epoch_time_df["avg_epoch_time"].min())))
    ax.set_ylim(lower_bound, upper_bound)

    # Remove minor ticks on x-axis and format new ticks
    ax.xaxis.set_minor_locator(NullLocator())
    ax.xaxis.set_major_formatter(ScalarFormatter())
    ax.set_xticks(num_gpus)

    ax.set_xlabel("Number of GPUs")
    ax.set_ylabel("Average Epoch Time (s)")
    ax.set_title("Average Epoch Time vs Number of GPUs")
    ax.legend(title="Method")

    num_datapoints = len(unique_nodes)
    figure_height, figure_width = calculate_plot_dimensions(num_datapoints=num_datapoints)
    fig.set_figheight(figure_height)
    fig.set_figwidth(figure_width)

    plt.tight_layout()
    sns.reset_orig()

    return fig, ax


def relative_epoch_time_speedup_plot(
    avg_epoch_time_df: pd.DataFrame, gpus_per_node: int = 4
) -> Tuple[Figure, Axes]:
    """Creates a log-log plot showing the relative training speedup for distributed
    training strategies as the number of GPUs increases.

    Args:
        avg_epoch_time_df (pd.DataFrame): A DataFrame containing the following columns:
            - "nodes": Number of nodes used in the training process.
            - "avg_epoch_time": Average time (in seconds) taken for an epoch.
            - "name": Name of the distributed training strategy.
        gpus_per_node (int): Number of GPUs per node. Used to calculate the total number
            of GPUs for each training configuration. Defaults to 4.

    Returns:
        Tuple[Figure, Axes]: A tuple containing the matplotlib `Figure` and `Axes` objects
        of the generated plot.

    Raises:
        ValueError: If `avg_epoch_time_df` is missing required columns.
    """
    sns.set_theme()
    fig, ax = plt.subplots(figsize=(6, 4))

    avg_epoch_time_df["num_gpus"] = avg_epoch_time_df["nodes"] * gpus_per_node
    avg_epoch_time_df["linear_speedup"] = avg_epoch_time_df["nodes"].astype(float)

    # Plotting the speedup for each strategy
    strategy_names = sorted(avg_epoch_time_df["name"].unique())
    for strategy in strategy_names:
        strategy_data = avg_epoch_time_df[avg_epoch_time_df.name == strategy]

        base_time = strategy_data["avg_epoch_time"].iloc[0]
        speedup = base_time / strategy_data["avg_epoch_time"]
        num_gpus = strategy_data["num_gpus"]

        marker = next(marker_cycle)
        ax.plot(num_gpus, speedup, marker=marker, lw=1.0, label=strategy, alpha=0.7)

    # Plotting the linear line
    num_gpus = np.array(avg_epoch_time_df["num_gpus"].unique())
    linear_speedup = np.array(avg_epoch_time_df["linear_speedup"].unique())
    ax.plot(num_gpus, linear_speedup, ls="dashed", lw=1.0, c="k", label="linear speedup")

    # The log scale must be set before changing the ticks
    ax.set_yscale("log", base=2)
    ax.set_xscale("log", base=2)

    # Making the numbers on the y-axis scalars instead of exponents
    ax.yaxis.set_major_formatter(ScalarFormatter())

    # Remove minor ticks on x-axis and format new ticks
    ax.xaxis.set_minor_locator(NullLocator())
    ax.xaxis.set_major_formatter(ScalarFormatter())
    ax.set_xticks(num_gpus)

    ax.set_xlabel("Number of GPUs")
    ax.set_ylabel("Speedup")
    ax.set_xticks(num_gpus)
    ax.set_title("Relative Speedup of Average Epoch Time vs Number of GPUs")
    ax.legend(ncol=1)

    num_datapoints = len(num_gpus)
    figure_height, figure_width = calculate_plot_dimensions(num_datapoints=num_datapoints)
    fig.set_figheight(figure_height)
    fig.set_figwidth(figure_width)

    plt.tight_layout()
    sns.reset_orig()

    return fig, ax


def gpu_bar_plot(data_df: pd.DataFrame, plot_title: str, y_label: str, main_column: str):
    """Creates a centered bar plot grouped by number of GPUs and strategy.

    Args:
        data_df (pd.DataFrame): DataFrame containing "strategy", "num_global_gpus",
            and `main_column`.
        plot_title (str): The title of the plot.
        y_label (str): The label for the y-axis.
        main_column (str): Column name for bar heights.

    Returns:
        Tuple[Figure, Axes]: The generated bar plot.
    """
    sns.set_theme()

    unique_gpu_counts = np.sort(data_df["num_global_gpus"].unique())

    # Deciding the color for each strategy in advance
    color_map = plt.get_cmap("tab10")
    strategy_color = {
        strategy: color_map(idx % 10)
        for idx, strategy in enumerate(data_df["strategy"].unique())
    }

    fig, ax = plt.subplots()
    x_positions = np.arange(len(unique_gpu_counts))  # Fixed positions for GPU counts

    # To store dynamic x offsets per GPU count
    bar_positions = {}

    # Calculating the global bar_width based on the highest number of adjacent bars
    max_num_strategies = data_df.groupby(["num_global_gpus"])["strategy"].nunique().max()
    bar_width = 1 / (max_num_strategies + 1)

    # To only add labels the first time we see a strategy
    seen_strategies = set()

    for i, gpu_count in enumerate(unique_gpu_counts):
        subset = data_df[data_df["num_global_gpus"] == gpu_count]
        strategies = subset["strategy"].unique()
        num_strategies = len(strategies)

        # Center the bars
        offset = (num_strategies - 1) / 2

        bar_positions[gpu_count] = {
            strategy: x_positions[i] + (idx - offset) * bar_width
            for idx, strategy in enumerate(strategies)
        }

        for strategy in strategies:
            strategy_data = subset[subset["strategy"] == strategy]
            color = strategy_color[strategy]

            label = strategy if strategy not in seen_strategies else None
            seen_strategies.add(strategy)
            ax.bar(
                x=bar_positions[gpu_count][strategy],
                height=strategy_data[main_column].values[0],
                width=bar_width,
                label=label,
                color=color,
            )

    ax.set_xlabel("Number of GPUs")
    ax.set_ylabel(y_label)
    ax.set_title(plot_title)
    ax.set_xticks(x_positions)
    ax.set_xticklabels(unique_gpu_counts)
    ax.legend()

    sns.reset_orig()

    return fig, ax


def computation_fraction_bar_plot(
    communication_data_df: pd.DataFrame,
) -> Tuple[Figure, Axes]:
    """Creates a stacked bar plot showing computation and communication fractions for
    different strategies and GPU counts.

    Args:
        communication_data_df (pd.DataFrame): A DataFrame containing the following columns:
            - "strategy": The name of the distributed training strategy.
            - "num_gpus": The number of GPUs used.
            - "computation_fraction": The fraction of time spent on computation.

    Returns:
        Tuple[Figure, Axes]: A tuple containing the matplotlib `Figure` and `Axes` objects
        of the generated plot.

    Raises:
        ValueError: If the DataFrame is missing required columns or has invalid data.
    """
    sns.set_theme()
    color_map = plt.get_cmap("tab10")
    hatch_patterns = ["//", r"\\"]

    strategy_labels = sorted(communication_data_df["strategy"].unique())
    gpu_numbers = sorted(communication_data_df["num_gpus"].unique())
    values = communication_data_df.pivot(
        index="strategy", columns="num_gpus", values="computation_fraction"
    ).to_numpy()

    width = 1 / (len(strategy_labels) + 1)
    complements = 1 - values

    x = np.arange(len(gpu_numbers))
    fig, ax = plt.subplots()

    # Creating an offset to "center" around zero
    static_offset = (len(strategy_labels) - 1) / 2
    for strategy_idx in range(len(strategy_labels)):
        dynamic_bar_offset = strategy_idx - static_offset

        color = color_map(strategy_idx % 10)
        hatch = hatch_patterns[strategy_idx % 2]

        ax.bar(
            x=x + dynamic_bar_offset * width,
            height=values[strategy_idx],
            width=width,
            color=color,
            label=strategy_labels[strategy_idx],
            edgecolor="gray",
            linewidth=0.6,
        )
        ax.bar(
            x=x + dynamic_bar_offset * width,
            height=complements[strategy_idx],
            width=width,
            bottom=values[strategy_idx],
            facecolor="none",
            edgecolor="gray",
            alpha=0.8,
            linewidth=0.6,
            hatch=hatch,
        )

    ax.set_ylabel("Computation fraction")
    ax.set_xlabel("Number of GPUs")
    ax.set_title("Computation vs Communication Time by Framework and Number of GPUs")
    ax.set_xticks(x)
    ax.set_xticklabels(gpu_numbers)
    ax.set_ylim(0, 1.1)

    # Adding communication time to the legend
    hatch_patch = Patch(facecolor="none", edgecolor="gray", hatch="//", label="Communication")
    ax.legend(handles=ax.get_legend_handles_labels()[0] + [hatch_patch])

    # Dynamically adjusting the width of the figure
    figure_width = max(int(2 * len(gpu_numbers)), 8)
    fig.set_figwidth(figure_width)
    fig.set_figheight(figure_width * 0.8)

    # Resetting so that seaborn's theme doesn't affect other plots
    sns.reset_orig()

    return fig, ax



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


import pytest
import ray.tune
from ray.tune.search.sample import Categorical, Float, Function, Integer

from itwinai.torch.tuning import (
    search_space,
)


def test_search_space_parsing():
    """Test search space config parser for all Ray Tune types and exceptions,
    when the configuration needs to be instantiated.
    """
    search_params = search_space(
        {
            "uniform_param": {"type": "uniform", "lower": 1e-5, "upper": 1e-3},
            "quniform_param": {"type": "quniform", "lower": 1, "upper": 10, "q": 0.5},
            "loguniform_param": {"type": "loguniform", "lower": 1, "upper": 100},
            "qloguniform_param": {"type": "qloguniform", "lower": 1, "upper": 100, "q": 0.5},
            "randn_param": {"type": "randn", "mean": 0.0, "sd": 1.0},
            "qrandn_param": {"type": "qrandn", "mean": 0.0, "sd": 1.0, "q": 0.5},
            "randint_param": {"type": "randint", "lower": 1, "upper": 100},
            "qrandint_param": {"type": "qrandint", "lower": 1, "upper": 100, "q": 5},
            "lograndint_param": {"type": "lograndint", "lower": 1, "upper": 100},
            "qlograndint_param": {"type": "qlograndint", "lower": 1, "upper": 100, "q": 5},
            "choice_param": {"type": "choice", "categories": [2, 4]},
            "grid_param": {"type": "grid_search", "values": [1, 2, 3]},
        }
    )
    assert isinstance(search_params, dict)
    assert isinstance(search_params["uniform_param"], Float)
    assert search_params["uniform_param"].lower == 1e-5
    assert search_params["uniform_param"].upper == 1e-3

    assert isinstance(search_params["quniform_param"], Float)
    assert search_params["quniform_param"].lower == 1
    assert search_params["quniform_param"].upper == 10

    assert isinstance(search_params["loguniform_param"], Float)
    assert search_params["loguniform_param"].lower == 1
    assert search_params["loguniform_param"].upper == 100

    assert isinstance(search_params["qloguniform_param"], Float)
    assert search_params["qloguniform_param"].lower == 1
    assert search_params["qloguniform_param"].upper == 100

    assert isinstance(search_params["randn_param"], Float)

    assert isinstance(search_params["qrandn_param"], Float)

    assert isinstance(search_params["randint_param"], Integer)
    assert search_params["randint_param"].lower == 1
    assert search_params["randint_param"].upper == 100

    assert isinstance(search_params["qrandint_param"], Integer)
    assert search_params["qrandint_param"].lower == 1
    assert search_params["qrandint_param"].upper == 100

    assert isinstance(search_params["lograndint_param"], Integer)
    assert search_params["lograndint_param"].lower == 1
    assert search_params["lograndint_param"].upper == 100

    assert isinstance(search_params["qlograndint_param"], Integer)
    assert search_params["qlograndint_param"].lower == 1
    assert search_params["qlograndint_param"].upper == 100

    assert isinstance(search_params["choice_param"], Categorical)
    assert search_params["choice_param"].categories == [2, 4]
    assert isinstance(search_params["grid_param"], dict)

    assert search_params["grid_param"] == {"grid_search": [1, 2, 3]}

    search_params = search_space(None)
    assert isinstance(search_params, dict)
    assert search_params == {}

    with pytest.raises(ValueError):
        search_space({"invalid": "value"})

    with pytest.raises(AttributeError):
        search_space({"invalid_param": {"type": "unknown_type"}})


def test_search_space_objects():
    """Test search space config parser for all Ray Tune types and exceptions,
    when the configuration doesn't need to be instantiated.
    """
    search_params = search_space(
        {
            "uniform_param": ray.tune.uniform(lower=1e-5, upper=1e-3),
            "quniform_param": ray.tune.quniform(lower=1, upper=10, q=0.5),
            "loguniform_param": ray.tune.loguniform(lower=1, upper=100),
            "qloguniform_param": ray.tune.qloguniform(lower=1, upper=100, q=0.5),
            "randn_param": ray.tune.randn(mean=0, sd=1),
            "qrandn_param": ray.tune.qrandn(mean=0, sd=1, q=0.5),
            "randint_param": ray.tune.randint(lower=1, upper=100),
            "qrandint_param": ray.tune.qrandint(lower=1, upper=100, q=5),
            "lograndint_param": ray.tune.lograndint(lower=1, upper=100),
            "qlograndint_param": ray.tune.qlograndint(lower=1, upper=100, q=5),
            "choice_param": ray.tune.choice(categories=[2, 4]),
            "grid_param": ray.tune.grid_search(values=[1, 2, 3]),
            "func_param": ray.tune.sample_from(lambda spec: spec.config.uniform_param * 0.01),
        }
    )
    assert isinstance(search_params, dict)
    assert isinstance(search_params["uniform_param"], Float)
    assert search_params["uniform_param"].lower == 1e-5
    assert search_params["uniform_param"].upper == 1e-3

    assert isinstance(search_params["quniform_param"], Float)
    assert search_params["quniform_param"].lower == 1
    assert search_params["quniform_param"].upper == 10

    assert isinstance(search_params["loguniform_param"], Float)
    assert search_params["loguniform_param"].lower == 1
    assert search_params["loguniform_param"].upper == 100

    assert isinstance(search_params["qloguniform_param"], Float)
    assert search_params["qloguniform_param"].lower == 1
    assert search_params["qloguniform_param"].upper == 100

    assert isinstance(search_params["randn_param"], Float)

    assert isinstance(search_params["qrandn_param"], Float)

    assert isinstance(search_params["randint_param"], Integer)
    assert search_params["randint_param"].lower == 1
    assert search_params["randint_param"].upper == 100

    assert isinstance(search_params["qrandint_param"], Integer)
    assert search_params["qrandint_param"].lower == 1
    assert search_params["qrandint_param"].upper == 100

    assert isinstance(search_params["lograndint_param"], Integer)
    assert search_params["lograndint_param"].lower == 1
    assert search_params["lograndint_param"].upper == 100

    assert isinstance(search_params["qlograndint_param"], Integer)
    assert search_params["qlograndint_param"].lower == 1
    assert search_params["qlograndint_param"].upper == 100

    assert isinstance(search_params["choice_param"], Categorical)
    assert search_params["choice_param"].categories == [2, 4]
    assert isinstance(search_params["grid_param"], dict)

    assert search_params["grid_param"] == {"grid_search": [1, 2, 3]}

    assert isinstance(search_params["func_param"], Function)

    search_params = search_space(None)
    assert isinstance(search_params, dict)
    assert search_params == {}

    with pytest.raises(ValueError):
        search_space({"invalid": "value"})

    with pytest.raises(AttributeError):
        search_space({"invalid_param": {"type": "unknown_type"}})



import argparse
import os
import sys
from pathlib import Path
from typing import Dict

import matplotlib.pyplot as plt
import ray
import torch
from ray import train, tune
from ray.tune import ResultGrid

from itwinai.parser import ConfigParser
from itwinai.utils import load_yaml

sys.path.append(os.path.join(os.path.dirname(__file__), 'src'))
sys.path.append(os.path.join(os.path.dirname(__file__), 'preprocessing'))


def run_trial(config: Dict, data: Dict):
    """Execute a single trial using the given configuration (config).
    This runs a full training pipeline - you can also specify a pipeline as a dictionary,
    e.g. if you only want to run certain parts without changing your config.yaml file
    (see below).

    Args:
        config (dict): A dictionary containing hyperparameters, such as:
            - 'batch_size' (int): The size of the batch for training.
            - 'lr' (float): The learning rate for the optimizer.
        data (dict): A dictionary containing a "pipeline_path" field, which points to the yaml
            file containing the pipeline definition
    """
    yaml_config = load_yaml('pipeline.yaml')

    seasons_list = yaml_config['seasons']
    for season in seasons_list:
        model_uri = f"outputs/cvae_model_{season}1d_1memb.pth"
        override_dict = {
            'season': season,
            'model_uri': model_uri,
            'pipeline.init_args.steps.training-step.init_args.logger': None,
            'pipeline.init_args.steps.evaluation-step.init_args.logger': None,
            'batch_size': config['batch_size'],
            'lr': config['lr']
        }

        parser = ConfigParser(config=yaml_config, override_keys=override_dict)
        my_pipeline = parser.parse_pipeline()
        print(f"Running pipeline for season: {season}")
        my_pipeline.execute()


def run_hpo(args: argparse.Namespace) -> None:
    """Run hyperparameter optimization using Ray Tune.
    Either starts a new optimization run or resumes from previous results.

    Args:
        args: Command-line arguments parsed by argparse.
    """
    if not args.load_old_results:

        # Initialize Ray with cluster configuration from environment variables
        ray.init(address="auto")

        # Define the search space for hyperparameters
        search_space = {
            'batch_size': tune.choice([64, 128]),
            'lr': tune.uniform(1e-5, 1e-3)
        }

        # TuneConfig for configuring search algorithm and scheduler
        tune_config = tune.TuneConfig(
            metric=args.metric,  # Metric to optimize (loss by default)
            mode="min",  # Minimize the loss
            num_samples=args.num_samples  # Number of trials to run
        )

        # Ray's RunConfig for experiment name and stopping criteria
        run_config = train.RunConfig(
            name="XTClim-Ray-Experiment", stop={"training_iteration": args.max_iterations}
        )

        # Determine GPU and CPU utilization per trial
        # We are allocating all available resources per node evenly across trials
        ngpus_per_trial = max(1, args.ngpus // args.num_samples)
        ncpus_per_trial = max(1, args.ncpus // args.num_samples)

        # Set up Ray Tune Tuner with resources and parameters
        resources_per_trial = {"gpu": ngpus_per_trial, "cpu": ncpus_per_trial}
        trainable_with_resources = tune.with_resources(
            run_trial,
            resources=resources_per_trial
        )

        data = {"pipeline_name": args.pipeline_name}
        trainable_with_parameters = tune.with_parameters(trainable_with_resources, data=data)

        tuner = tune.Tuner(
            trainable_with_parameters,
            tune_config=tune_config,
            run_config=run_config,
            param_space=search_space
        )

        # Run the hyperparameter optimization and get results
        result_grid = tuner.fit()

    else:
        # Load results from an earlier Ray Tune run
        print(f"Loading results from {args.experiment_path}...")

        # Restore tuner from saved results
        restored_tuner = tune.Tuner.restore(args.experiment_path, trainable=run_trial)
        result_grid = restored_tuner.get_results()

    # Display experiment statistics
    print(f"Number of errored trials: {result_grid.num_errors}")
    print(f"Number of terminated trials: {result_grid.num_terminated}")
    print(f"Ray Tune experiment path: {result_grid.experiment_path}")

    # Get the best result based on the last 10 iterations' average
    best_result = result_grid.get_best_result(
        scope="last-10-avg",
        metric=args.metric,
        mode="min"
    )
    print(f"Best result: {best_result}")

    # Print a dataframe with all trial results
    result_df = result_grid.get_dataframe()
    print(f"All results dataframe: {result_df}")
    print(f"All result columns: {result_df.columns}")

    # Plot the results for all trials
    plot_results(result_grid, metric=args.metric, filename="ray-loss-plot.png")
    plot_results(result_grid, metric="valid_loss", filename="ray-valid_loss-plot.png")


def plot_results(result_grid: ResultGrid, metric: str = "loss", filename: str = "plot.png"):
    """Plot the results for all trials and save the plot to a file.

    Args:
        result_grid (ResultGrid): Results from Ray Tune trials.
        metric (str): The metric to plot (e.g., 'loss').
        filename (str): Name of the file to save the plot.
    """
    fig, ax = plt.subplots()

    for result in result_grid:
        label = f"lr={result.config['lr']:.6f}, batch size={result.config['batch_size']}"
        result.metrics_dataframe.plot("training_iteration", metric, ax=ax, label=label)

    ax.set_title(f"{metric.capitalize()} vs. Training Iteration for All Trials")
    ax.set_ylabel(metric.capitalize())

    fig.savefig(filename)

    # Show the plot
    plt.show()


# Main entry point for script execution
if __name__ == "__main__":

    # Parse command-line arguments
    parser = argparse.ArgumentParser(
        description='Hyperparameter Optimization with Ray Tune'
    )
    parser.add_argument(
        '--load_old_results',
        type=bool,
        default=False,
        help='Set this to true if you want to load results from an older ray run.'
    )
    parser.add_argument(
        '--pipeline_name',
        type=str,
        default='training_pipeline',
        help='Name of the training pipeline to be used. \
            This pipeline has to be defined in a file called "config.yaml". \
            Defaults to "training_pipeline"'
    )
    parser.add_argument(
        '--experiment_path',
        type=str,
        default='~/ray_results/XTClim-Ray-Experiment',
        help='Directory where the results of the previous run are stored. \
        Set this only if load_old_results is set to True. \
        Defaults to ~/ray_results/XTClim-Ray-Experiment'
    )
    parser.add_argument('--num_samples', type=int, default=10, help='Number of trials to run')
    parser.add_argument('--ngpus', type=int, help='Number of GPUs available on node.')
    parser.add_argument('--ncpus', type=int, help='Number of CPUs available on node.')
    parser.add_argument('--metric', type=str, default='loss', help='Metric to optimise.')
    parser.add_argument('--max_iterations', type=int, default='20', help='Maximum iterations per trial')

    args = parser.parse_args()  # Parse the command-line arguments

    run_hpo(args)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Utilities for itwinai package."""

import inspect
import os
import random
import sys
from collections.abc import MutableMapping
from pathlib import Path
from typing import Callable, Dict, Hashable, List, Tuple, Type
from urllib.parse import urlparse

import yaml

adjectives = [
    "quantum",
    "relativistic",
    "wavy",
    "entangled",
    "chiral",
    "tachyonic",
    "superluminal",
    "anomalous",
    "hypercharged",
    "fermionic",
    "hadronic",
    "quarky",
    "holographic",
    "dark",
    "force-sensitive",
    "chaotic",
]

names = [
    "neutrino",
    "graviton",
    "muon",
    "gluon",
    "tachyon",
    "quasar",
    "pulsar",
    "blazar",
    "meson",
    "boson",
    "hyperon",
    "starlord",
    "groot",
    "rocket",
    "yoda",
    "skywalker",
    "sithlord",
    "midichlorian",
    "womp-rat",
    "beskar",
    "mandalorian",
    "ewok",
    "vibranium",
    "nova",
    "gamora",
    "drax",
    "ronan",
    "thanos",
    "cosmo",
]


def generate_random_name():
    return f"{random.choice(adjectives)}-{random.choice(names)}"


def load_yaml(path: str) -> Dict:
    """Load YAML file as dict.

    Args:
        path (str): path to YAML file.

    Raises:
        yaml.YAMLError: for loading/parsing errors.

    Returns:
        Dict: nested dict representation of parsed YAML file.
    """
    with open(path, "r", encoding="utf-8") as yaml_file:
        try:
            loaded_config = yaml.safe_load(yaml_file)
        except yaml.YAMLError as exc:
            print(exc)
            raise exc
    return loaded_config


def dynamically_import_class(name: str) -> Type:
    """
    Dynamically import class by module path.
    Adapted from https://stackoverflow.com/a/547867

    Args:
        name (str): path to the class (e.g., mypackage.mymodule.MyClass)

    Returns:
        __class__: class type.
    """
    try:
        module, class_name = name.rsplit(".", 1)
        mod = __import__(module, fromlist=[class_name])
        klass = getattr(mod, class_name)
    except ModuleNotFoundError as err:
        print(
            f"Module not found when trying to dynamically import '{name}'. "
            "Make sure that the module's file is reachable from your current "
            "directory."
        )
        raise err
    except Exception as err:
        print(
            f"Exception occurred when trying to dynamically import '{name}'. "
            "Make sure that the module's file is reachable from your current "
            "directory and that the class is present in that module."
        )
        raise err

    return klass


def flatten_dict(d: MutableMapping, parent_key: str = "", sep: str = ".") -> MutableMapping:
    """Flatten dictionary

    Args:
        d (MutableMapping): nested dictionary to flatten
        parent_key (str, optional): prefix for all keys. Defaults to ''.
        sep (str, optional): separator for nested key concatenation.
            Defaults to '.'.

    Returns:
        MutableMapping: flattened dictionary with new keys.
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class SignatureInspector:
    """Provides the functionalities to inspect the signature of a function
    or a method.

    Args:
        func (Callable): function to be inspected.
    """

    INFTY: int = sys.maxsize

    def __init__(self, func: Callable) -> None:
        self.func = func
        self.func_params = inspect.signature(func).parameters.items()

    @property
    def has_varargs(self) -> bool:
        """Checks if the function has ``*args`` parameter."""
        return any(map(lambda p: p[1].kind == p[1].VAR_POSITIONAL, self.func_params))

    @property
    def has_kwargs(self) -> bool:
        """Checks if the function has ``**kwargs`` parameter."""
        return any(map(lambda p: p[1].kind == p[1].VAR_KEYWORD, self.func_params))

    @property
    def required_params(self) -> Tuple[str]:
        """Names of required parameters. Class method's 'self' is skipped."""
        required_params = list(
            filter(
                lambda p: (
                    p[0] != "self"
                    and p[1].default == inspect._empty
                    and p[1].kind != p[1].VAR_POSITIONAL
                    and p[1].kind != p[1].VAR_KEYWORD
                ),
                self.func_params,
            )
        )
        return tuple(map(lambda p: p[0], required_params))

    @property
    def min_params_num(self) -> int:
        """Minimum number of arguments required."""
        return len(self.required_params)

    @property
    def max_params_num(self) -> int:
        """Max number of supported input arguments.
        If no limit, ``SignatureInspector.INFTY`` is returned.
        """
        if self.has_kwargs or self.has_varargs:
            return self.INFTY
        return len(self.func_params)


def clear_key(my_dict: Dict, dict_name: str, key: Hashable, complain: bool = True) -> Dict:
    """Remove key from dictionary if present and complain.

    Args:
        my_dict (Dict): Dictionary.
        dict_name (str): name of the dictionary.
        key (Hashable): Key to remove.
    """
    if key in my_dict:
        if complain:
            print(f"Field '{key}' should not be present in dictionary '{dict_name}'")
        del my_dict[key]
    return my_dict


def make_config_paths_absolute(args: List[str]):
    """Process CLI arguments to make paths specified for `--config-path` or `-cp` absolute.
    Returns the modified arguments list.

    Args:
        args (List[str]): a list of system arguments

    Returns:
        List(str): the updated list of system arguments, where all the config path argument is
            absolute.
    """
    updated_args = args.copy()
    for i, arg in enumerate(updated_args):
        if arg.startswith("--config-path=") or arg.startswith("-cp="):
            prefix, path = arg.split("=", 1)
            abs_path = os.path.abspath(path)
            updated_args[i] = f"{prefix}={abs_path}"
            sys.path.append(abs_path)
            break
        elif arg in {"--config-path", "-cp"}:
            # Handle the case where the path is in the next argument
            abs_path = os.path.abspath(updated_args[i + 1])
            updated_args[i + 1] = abs_path
            sys.path.append(abs_path)
            break
    return updated_args


def get_root_cause(exception: Exception) -> Exception:
    """Recursively extract the first exception in the exception chain."""
    root = exception
    while root.__cause__ is not None:  # Traverse the exception chain
        root = root.__cause__
    return root


def to_uri(path_str: str | Path) -> str:
    """Parse a path and convert it to a URI.

    Args:
        path_str (str): path to convert.

    Returns:
        str: URI.
    """
    if isinstance(path_str, Path):
        return str(Path(path_str).resolve())
    parsed = urlparse(path_str)
    if parsed.scheme:
        # If it has a scheme, assume it's a URI and return as-is
        return path_str
    # Otherwise, make it absolute
    return str(Path(path_str).resolve())



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Tests for CERN use case (3DGAN)."""

import os
import subprocess
import tempfile
from pathlib import Path

import pytest

CERN_PATH = Path("use-cases", "3dgan").resolve()
CKPT_NAME = "3dgan-inference.pth"
DEFAULT_DATASET_PATH = "exp_data"


@pytest.mark.skip("deprecated")
def test_structure_3dgan(check_folder_structure):
    """Test 3DGAN folder structure."""
    check_folder_structure(CERN_PATH)


@pytest.mark.functional
def test_3dgan_train(torch_env, install_requirements, tmp_path):
    """Test 3DGAN torch lightning trainer by running it end-to-end.

    If CERN_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(CERN_PATH, torch_env)
    dataset_path = os.environ.get("CERN_DATASET", DEFAULT_DATASET_PATH)
    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config-path {CERN_PATH} "
        f"dataset_location={dataset_path} "
        "hw_accelerators=auto "
        "distributed_strategy=auto "
        "mlflow_tracking_uri=ml_logs/mlflow"
    )

    subprocess.run(cmd.split(), check=True, cwd=tmp_path)


@pytest.mark.functional
def test_3dgan_inference(
    torch_env,
    install_requirements,
    tmp_path,
    # fake_model_checkpoint
):
    """Test 3DGAN torch lightning trainer by running it end-to-end.

    If CERN_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(CERN_PATH, torch_env)

    # Test inference
    dataset_path = os.environ.get("CERN_DATASET", DEFAULT_DATASET_PATH)

    exec = CERN_PATH / "create_inference_sample.py"

    run_inference_cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config-path {CERN_PATH} +pipe_key=inference_pipeline "
        f"dataset_location={dataset_path} "
        f"inference_model_uri={CKPT_NAME} "
        "hw_accelerators=auto "
        "distributed_strategy=auto "
        "logs_dir=ml_logs/mlflow_logs "
        "inference_results_location=3dgan-generated-data "
    )

    # Create fake inference dataset and checkpoint
    generate_model_cmd = (
        f"{torch_env}/bin/python {exec} "
        f"--root {str(Path(tmp_path).resolve())} "
        f"--ckpt-name {CKPT_NAME}"
    )
    subprocess.run(generate_model_cmd.split(), check=True, cwd=tmp_path)

    # Running the inference
    subprocess.run(run_inference_cmd.split(), check=True, cwd=tmp_path)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Provide functionalities to manage configuration files, including parsing,
execution, and dynamic override of fields.
"""

import logging
import os
from typing import List, Optional, Type, Union

from jsonargparse import ActionConfigFile
from jsonargparse import ArgumentParser as JAPArgumentParser
from jsonargparse._formatters import DefaultHelpFormatter


class ArgumentParser(JAPArgumentParser):
    """Wrapper of ``jsonargparse.ArgumentParser``.
    Initializer for ArgumentParser instance. It can parse arguments from
    a series of configuration files. Example:

    >>> python main.py --config base-conf.yaml --config other-conf.yaml \\
    >>> --param OVERRIDE_VAL

    All the arguments from the initializer of `argparse.ArgumentParser
    <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
    are supported. Additionally it accepts:

    Args:
        env_prefix (Union[bool, str], optional): Prefix for environment
            variables. ``True`` to derive from ``prog``.. Defaults to True.
        formatter_class (Type[DefaultHelpFormatter], optional): Class for
            printing help messages. Defaults to DefaultHelpFormatter.
        exit_on_error (bool, optional): Defaults to True.
        logger (Union[bool, str, dict, logging.Logger], optional):
            Configures the logger, see :class:`.LoggerProperty`.
            Defaults to False.
        version (Optional[str], optional): Program version which will be
            printed by the --version argument. Defaults to None.
        print_config (Optional[str], optional): Add this as argument to
            print config, set None to disable.
            Defaults to "--print_config".
        parser_mode (str, optional): Mode for parsing config files:
            ``'yaml'``, ``'jsonnet'`` or ones added via
            :func:`.set_loader`.. Defaults to "yaml".
        dump_header (Optional[List[str]], optional): Header to include
            as comment when dumping a config object. Defaults to None.
        default_config_files
            (Optional[List[Union[str, os.PathLike]]], optional):
            Default config file locations, e.g.
            :code:`['~/.config/myapp/*.yaml']`. Defaults to None.
        default_env (bool, optional): Set the default value on whether
            to parse environment variables. Defaults to False.
        default_meta (bool, optional): Set the default value on whether
            to include metadata in config objects. Defaults to True.
    """

    def __init__(
        self,
        *args,
        env_prefix: Union[bool, str] = True,
        formatter_class: Type[DefaultHelpFormatter] = DefaultHelpFormatter,
        exit_on_error: bool = True,
        logger: Union[bool, str, dict, logging.Logger] = False,
        version: Optional[str] = None,
        print_config: Optional[str] = "--print_config",
        parser_mode: str = "yaml",
        dump_header: Optional[List[str]] = None,
        default_config_files: Optional[List[Union[str, os.PathLike]]] = None,
        default_env: bool = False,
        default_meta: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            env_prefix=env_prefix,
            formatter_class=formatter_class,
            exit_on_error=exit_on_error,
            logger=logger,
            version=version,
            print_config=print_config,
            parser_mode=parser_mode,
            dump_header=dump_header,
            default_config_files=default_config_files,
            default_env=default_env,
            default_meta=default_meta,
            **kwargs,
        )
        self.add_argument(
            "-c",
            "--config",
            action=ActionConfigFile,
            help="Path to a configuration file in json or yaml format.",
        )



"""
Train file to launch pipeline
"""

import os
import sys
from itwinai.parser import ConfigParser
from itwinai.utils import load_yaml

sys.path.append(os.path.join(os.path.dirname(__file__), "src"))
sys.path.append(os.path.join(os.path.dirname(__file__), "preprocessing"))


if __name__ == "__main__":

    config = load_yaml('pipeline.yaml')
    seasons_list = config['seasons']

    for season in seasons_list:
        model_uri = f"outputs/cvae_model_{season}1d_1memb.pth"
        override_dict = {
            'season': season,
            'model_uri': model_uri
        }
        pipe_parser = ConfigParser(
            config=config,
            override_keys=override_dict
        )
        pipeline = pipe_parser.parse_pipeline()

        print(f"Running pipeline for season: {season}")
        pipeline.execute()


# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Linus Eickhoff <linus.maximilian.eickhoff@cern.ch> - CERN
# --------------------------------------------------------------------------------------


"""Provides training logic for PyTorch models via Trainer classes."""

import logging
import os
import sys
import tempfile
import time
from collections import defaultdict
from pathlib import Path
from time import perf_counter as default_timer
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import ray.train
import ray.train.torch
import ray.tune
import torch
import torch.distributed as dist
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
import yaml
from ray.train import Checkpoint, DataConfig, RunConfig, ScalingConfig
from ray.train.torch import TorchConfig
from ray.train.torch import TorchTrainer as RayTorchTrainer
from ray.tune import TuneConfig
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.optim.lr_scheduler import LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm

# Cyclic imports...
from itwinai.torch.monitoring.monitoring import measure_gpu_utilization
from itwinai.torch.profiling.profiler import profile_torch_trainer

from ..components import Trainer, monitor_exec
from ..distributed import ray_cluster_is_running
from ..loggers import EpochTimeTracker, Logger, LogMixin
from ..utils import generate_random_name, load_yaml, to_uri
from .config import TrainingConfiguration
from .distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    NonDistributedStrategy,
    RayDDPStrategy,
    RayDeepSpeedStrategy,
    RayHorovodStrategy,
    RayTorchDistributedStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
    distributed_resources_available,
)
from .reproducibility import seed_worker, set_seed
from .tuning import search_space
from .type import Batch, Metric

if TYPE_CHECKING:
    from ray.train.horovod import HorovodConfig

py_logger = logging.getLogger(__name__)


def _get_tuning_metric_name(tune_config: TuneConfig | None) -> str:
    """Extracts the metric name from TuneConfig or scheduler in a generic way."""
    DEFAULT_NAME = "loss"

    if not tune_config:
        return DEFAULT_NAME

    # Try to get from TuneConfig
    if tune_config.metric:
        return tune_config.metric

    # Try to get from the scheduler (if defined)
    scheduler = tune_config.scheduler
    if scheduler and hasattr(scheduler, "metric") and scheduler.metric:
        return scheduler.metric

    return DEFAULT_NAME


class TorchTrainer(Trainer, LogMixin):
    """Trainer class for torch training algorithms.

    Args:
        config (Dict | TrainingConfiguration): training configuration
            containing hyperparameters.
        epochs (int): number of training epochs.
        model (Union[nn.Module, str] | None, optional): pytorch model to
            train or a string identifier. Defaults to None.
        strategy (Literal['ddp', 'deepspeed', 'horovod'], optional):
            distributed strategy. Defaults to 'ddp'.
        test_every (int | None, optional): run a test epoch
            every ``test_every`` epochs. Disabled if None. Defaults to None.
        random_seed (int | None, optional): set random seed for
            reproducibility. If None, the seed is not set. Defaults to None.
        logger (Logger | None, optional): logger for ML tracking.
            Defaults to None.
        metrics (Dict[str, Callable] | None, optional): map of torchmetrics
            metrics. Defaults to None.
        checkpoints_location (str): path to checkpoints directory.
            Defaults to "checkpoints".
        checkpoint_every (int | None): save a checkpoint every
            ``checkpoint_every`` epochs. Disabled if None. Defaults to None.
        disable_tqdm (bool): whether to disable tqdm progress bar(s).
        name (str | None, optional): trainer custom name. Defaults to None.
        profiling_wait_epochs (int): how many epochs to wait before starting
            the profiler.
        profiling_warmup_epochs (int): length of the profiler warmup phase in terms of
            number of epochs.
        measure_gpu_data (bool): enable the collection of data on average GPU utilization and
            total energy consumption throughout training. Defaults to False.
        measure_communication_overhead (bool): enable the profiling of computation and
            multi-worker communication operations. It uses the torch profiler and it may
            slow down training. Dafults to False.
        measure_epoch_time (bool): enable the measurement of epoch duration (in seconds).
            Defaults to False,
        ray_scaling_config (ScalingConfig, optional): scaling config for Ray Trainer.
            Defaults to None,
        ray_tune_config (TuneConfig, optional): tune config for Ray Tuner.
            Defaults to None.
        ray_run_config (RunConfig, optional): run config for Ray Trainer.
            Defaults to None.
        ray_search_space (Dict[str, Any], optional): search space for Ray Tuner.
            Defaults to None.
        ray_torch_config (TorchConfig, optional): torch configuration for Ray's TorchTrainer.
            Defaults to None.
        ray_data_config (DataConfig, optional): dataset configuration for Ray.
            Defaults to None.
        ray_horovod_config (HorovodConfig, optional): horovod configuration for Ray's
            HorovodTrainer. Defaults to None.
        from_checkpoint (str | Path, optional): path to checkpoint directory. Defaults to None.
        initial_best_validation_metric (str): initial value for the best validation metric.
            Usually the validation metric is a loss to be minimized and this value exceeds the
            highest possible loss value, so that it will be overwritten when the first
            vaidation loss is computed. Example values are "inf" and "-inf", depending on
            wether the best validation metric should be minimized or maximized.
            Defaults to "inf".
        run_id (str, optional): name used to identify a specific run when collecting
            metrics on the trainer (e.g. GPU utilization). Defaults to None.
        time_ray (bool): whether to time and log the execution of Ray functions.
            Defaults to False.
    """

    _strategy: TorchDistributedStrategy | None = None

    #: PyTorch ``DataLoader`` for training dataset.
    train_dataloader: DataLoader | None = None
    #: PyTorch ``DataLoader`` for validation dataset.
    validation_dataloader: DataLoader | None = None
    #: PyTorch ``DataLoader`` for test dataset.
    test_dataloader: DataLoader | None = None
    #: PyTorch model to train.
    model: nn.Module | None = None
    #: Loss criterion.
    loss: Callable | None = None
    #: Optimizer.
    optimizer: Optimizer | None = None
    #: Learning rate scheduler.
    lr_scheduler: LRScheduler | None = None
    #: PyTorch random number generator (PRNG).
    torch_rng: torch.Generator | None = None
    #: itwinai ``itwinai.Logger``
    logger: Logger | None = None
    #: Total number training batches used so far, across all epochs.
    train_glob_step: int = 0
    #: Total number validation batches used so far, across all epochs.
    validation_glob_step: int = 0
    #: Total number test batches used so far, across all epochs.
    test_glob_step: int = 0
    #: Dictionary of ``torchmetrics`` metrics, indexed by user-defined names.
    metrics: Dict[str, Callable]
    #: PyTorch Profiler for communication vs. computation comparison
    profiler: Any | None
    #: Toggle for GPU utilization monitoring
    measure_gpu_data: bool = False
    #: Toggle for communication vs computation fraction profiling
    measure_communication_overhead: bool = False
    #: Toggle for epoch time tracking
    measure_epoch_time: bool = False
    #: Run ID
    run_id: str
    #: Toggle for Ray time logging
    time_ray: bool = False

    def __init__(
        self,
        config: Dict | TrainingConfiguration,
        epochs: int,
        model: Union[nn.Module, str] | None = None,
        strategy: Literal["ddp", "deepspeed", "horovod"] | None = "ddp",
        test_every: int | None = None,
        random_seed: int | None = None,
        logger: Logger | None = None,
        metrics: Dict[str, Metric] | None = None,
        checkpoints_location: str | Path = "checkpoints",
        checkpoint_every: int | None = None,
        disable_tqdm: bool = False,
        name: str | None = None,
        profiling_wait_epochs: int = 1,
        profiling_warmup_epochs: int = 2,
        measure_gpu_data: bool = False,
        measure_communication_overhead: bool = False,
        measure_epoch_time: bool = False,
        ray_scaling_config: ScalingConfig | None = None,
        ray_tune_config: TuneConfig | None = None,
        ray_run_config: RunConfig | None = None,
        ray_search_space: Dict[str, Any] | None = None,
        ray_torch_config: TorchConfig | None = None,
        ray_data_config: DataConfig | None = None,
        ray_horovod_config: Optional["HorovodConfig"] = None,
        from_checkpoint: str | Path | None = None,
        initial_best_validation_metric: str = "inf",
        run_id: str | None = None,
        time_ray: bool = False,
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))

        # config is mean to store all hyperparameters, which can very from use
        # case to use case and include learning_rate, batch_size....
        config = {} if config is None else config
        if isinstance(config, dict):
            config = TrainingConfiguration(**config)

        self.config = config
        self.epochs = epochs
        self.model = model
        self.strategy = strategy
        self.test_every = test_every
        self.random_seed = random_seed
        self.logger = logger
        self.metrics = metrics if metrics is not None else {}
        self.checkpoints_location = checkpoints_location
        os.makedirs(self.checkpoints_location, exist_ok=True)
        self.checkpoint_every = checkpoint_every
        self.disable_tqdm = disable_tqdm
        self.profiler = None
        self.profiling_wait_epochs = profiling_wait_epochs
        self.profiling_warmup_epochs = profiling_warmup_epochs
        self.measure_gpu_data = measure_gpu_data
        self.measure_communication_overhead = measure_communication_overhead
        self.measure_epoch_time = measure_epoch_time
        self.ray_scaling_config = ray_scaling_config
        self.ray_tune_config = ray_tune_config
        self.ray_run_config = ray_run_config
        self.ray_search_space = ray_search_space
        self.ray_horovod_config = ray_horovod_config
        self.ray_torch_config = ray_torch_config
        self.ray_data_config = ray_data_config
        self.from_checkpoint = from_checkpoint
        self.time_ray = time_ray
        if self.from_checkpoint:
            self.from_checkpoint = Path(from_checkpoint)
            if not self.from_checkpoint.exists():
                raise RuntimeError(
                    "from_checkpoint argument was passed, but the checkpoint is not found "
                    f"at {self.from_checkpoint}"
                )

        if self.checkpoints_location:
            Path(self.checkpoints_location).mkdir(exist_ok=True, parents=True)

        py_logger.debug(f"ray_scaling_config: {ray_scaling_config}")
        py_logger.debug(f"ray_tune_config: {ray_tune_config}")
        py_logger.debug(f"ray_run_config: {ray_run_config}")
        py_logger.debug(f"ray_horovod_config: {ray_horovod_config}")
        py_logger.debug(f"ray_torch_config: {ray_torch_config}")
        py_logger.debug(f"ray_data_config: {ray_data_config}")
        py_logger.debug(f"ray_search_space: {ray_search_space}")

        # Initial training state -- can be resumed from a checkpoint
        self.model_state_dict = None
        self.optimizer_state_dict = None
        self.lr_scheduler_state_dict = None
        self.torch_rng_state = None
        # This is initialized to inf as it usually represents a loss to minimize.
        # If the validation metric is meant to be maximized, change this to -inf.
        self.best_validation_metric = float(initial_best_validation_metric)
        self.current_epoch = 0
        if run_id is None:
            run_id = generate_random_name()
        self.run_id = run_id

    @property
    def strategy(self) -> TorchDistributedStrategy:
        """Strategy currently in use."""
        return self._strategy

    @strategy.setter
    def strategy(self, strategy: str | TorchDistributedStrategy) -> None:
        if isinstance(strategy, TorchDistributedStrategy):
            self._strategy = strategy
        else:
            self._strategy = self._detect_distributed_strategy(strategy)

    @property
    def device(self) -> str:
        """Current device from distributed strategy."""
        return self.strategy.device()

    def _detect_distributed_strategy(self, strategy: str) -> TorchDistributedStrategy:
        """When a Ray cluster is detected the Ray-equivalent distributed strategy is
        automatically selected, without needing the user to explicitly set it.
        """

        py_logger.debug(f"Strategy was set to {strategy}")

        enough_resources = distributed_resources_available() or ray_cluster_is_running()
        py_logger.debug(
            f"Enough resources? {enough_resources} "
            f"(distributed_resources_available: {distributed_resources_available()}) "
            f"(ray_cluster_is_running: {ray_cluster_is_running()})"
        )

        # NOTE: setting strategy to None prevents the trainer to run distribtued ML, regardless
        # of the availability of the resources.
        if strategy is None or not enough_resources:
            py_logger.warning("Falling back to non-distributed strategy.")
            return NonDistributedStrategy()

        if ray_cluster_is_running():
            py_logger.info(
                f"Ray cluster was detected, thus the Ray equivalent for {strategy} is used"
            )

        match strategy, ray_cluster_is_running():
            case "ddp", True:
                return RayDDPStrategy()

            case "ddp", False:
                return TorchDDPStrategy(backend=self.config.dist_backend)

            case "horovod", True:
                return RayHorovodStrategy()

            case "horovod", False:
                return HorovodStrategy()

            case "deepspeed", True:
                return RayDeepSpeedStrategy(backend=self.config.dist_backend)

            case "deepspeed", False:
                return DeepSpeedStrategy(backend=self.config.dist_backend)

            case _:
                raise RuntimeError(f"Strategy '{strategy}' is not recognized.")

    def _init_distributed_strategy(self) -> None:
        if not self.strategy.is_initialized:
            self.strategy.init()

    def _set_optimizer_from_config(self) -> None:
        match self.config.optimizer:
            case "adadelta":
                self.optimizer = optim.Adadelta(
                    self.model.parameters(),
                    lr=self.config.optim_lr,
                    weight_decay=self.config.optim_weight_decay,
                )
            case "adam":
                self.optimizer = optim.Adam(
                    self.model.parameters(),
                    lr=self.config.optim_lr,
                    betas=self.config.optim_betas,
                    weight_decay=self.config.optim_weight_decay,
                )
            case "adamw":
                self.optimizer = optim.AdamW(
                    self.model.parameters(),
                    lr=self.config.optim_lr,
                    betas=self.config.optim_betas,
                    weight_decay=self.config.optim_weight_decay,
                )
            case "rmsprop":
                self.optimizer = optim.RMSprop(
                    self.model.parameters(),
                    lr=self.config.optim_lr,
                    weight_decay=self.config.optim_weight_decay,
                    momentum=self.config.optim_momentum,
                )
            case "sgd":
                self.optimizer = optim.SGD(
                    self.model.parameters(),
                    lr=self.config.optim_lr,
                    weight_decay=self.config.optim_weight_decay,
                    momentum=self.config.optim_momentum,
                )
            case _:
                raise ValueError(
                    "Unrecognized self.config.optimizer! Check the docs for "
                    "supported values and consider overriding "
                    "create_model_loss_optimizer method for more flexibility."
                )

    def _set_lr_scheduler_from_config(self) -> None:
        """Parse Lr scheduler from training config"""
        if not self.config.lr_scheduler:
            return
        if not self.optimizer:
            raise ValueError("Trying to instantiate a LR scheduler but the optimizer is None!")

        match self.config.lr_scheduler:
            case "constant":
                self.lr_scheduler = lr_scheduler.ConstantLR(self.optimizer)
            case "polynomial":
                self.lr_scheduler = lr_scheduler.PolynomialLR(self.optimizer)
            case "exponential":
                self.lr_scheduler = lr_scheduler.ExponentialLR(
                    self.optimizer, gamma=self.config.lr_scheduler_gamma
                )
            case "linear":
                self.lr_scheduler = lr_scheduler.LinearLR(self.optimizer)
            case "multistep":
                self.lr_scheduler = lr_scheduler.MultiStepLR(
                    self.optimizer, milestones=self.config.lr_scheduler_step_size
                )
            case "step":
                self.lr_scheduler = lr_scheduler.StepLR(
                    self.optimizer, step_size=self.config.lr_scheduler_step_size
                )
            case _:
                raise ValueError(
                    "Unrecognized self.config.lr_scheduler! Check the docs for "
                    "supported values and consider overriding "
                    "create_model_loss_optimizer method for more flexibility."
                )

    def _set_loss_from_config(self) -> None:
        match self.config.loss:
            case "nllloss":
                self.loss = nn.functional.nll_loss
            case "cross_entropy":
                self.loss = nn.functional.cross_entropy
            case "mse":
                self.loss = nn.functional.mse_loss
            case "bceloss":
                self.loss = nn.functional.binary_cross_entropy
            case _:
                raise ValueError(
                    "Unrecognized self.config.loss! Check the docs for "
                    "supported values and consider overriding "
                    "create_model_loss_optimizer method for more flexibility."
                )

    def _time_and_log(self, fn: Callable, identifier: str, step: int | None = None) -> Any:
        """Time and log the execution of a function (using time.monotonic()).

        Args:
            fn (Callable): function to execute, time and log (pass args using lambda)
            identifier (str): identifier for the logged metric
            step (int | None): step for logging

        Returns:
            result (Any): result of the function call
        """
        if not self.logger:
            py_logger.warning(f"No logger set! Cannot log time for {identifier}! ")
            return fn()

        if not self.logger.is_initialized:
            py_logger.warning(
                f"Logger context not initialized for timing {identifier}. Setting context."
            )
            self.logger.create_logger_context()

        step = step or self.current_epoch
        if step is None:
            py_logger.warning("current_epoch is not set and no explicit step was provided!")

        # Use monotonic time to avoid time drift
        t_start = time.monotonic()
        result = fn()
        t_end = time.monotonic()
        # already in seconds
        t_delta = t_end - t_start

        self.log(
            item=t_delta,
            identifier=identifier,
            kind="metric",
            step=step,
        )

        return result

    def get_default_distributed_kwargs(self) -> Dict:
        """Gives the default kwargs for the trainer's strategy's distributed() method."""

        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        elif isinstance(self.strategy, HorovodStrategy):
            import horovod.torch as hvd

            distribute_kwargs = dict(
                compression=(
                    hvd.Compression.fp16
                    if self.config.fp16_allreduce
                    else hvd.Compression.none
                ),
                op=hvd.Adasum if self.config.use_adasum else hvd.Average,
                gradient_predivide_factor=self.config.gradient_predivide_factor,
            )
        else:
            distribute_kwargs = {}

        return distribute_kwargs

    def create_model_loss_optimizer(self) -> None:
        """Instantiate a torch model, loss, optimizer, and LR scheduler using the
        configuration provided in the Trainer constructor.
        Generally a user-defined method.
        """
        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################

        # Model, optimizer, and lr scheduler may have already been loaded from a checkpoint

        if self.model is None:
            raise ValueError(
                "self.model is None! Either pass it to the constructor, load a checkpoint, or "
                "override create_model_loss_optimizer method."
            )
        if self.model_state_dict:
            # Load model from checkpoint
            self.model.load_state_dict(self.model_state_dict, strict=False)

        # Parse optimizer from training configuration
        # Optimizer can be changed with a custom one here!
        self._set_optimizer_from_config()

        # Parse LR scheduler from training configuration
        # LR scheduler can be changed with a custom one here!
        self._set_lr_scheduler_from_config()

        if self.optimizer_state_dict:
            # Load optimizer state from checkpoint
            # IMPORTANT: this must be after the learning rate scheduler was already initialized
            # by passing to it the optimizer. Otherwise the optimizer state just loaded will
            # be modified by the lr scheduler.
            self.optimizer.load_state_dict(self.optimizer_state_dict)

        if self.lr_scheduler_state_dict and self.lr_scheduler:
            # Load LR scheduler state from checkpoint
            self.lr_scheduler.load_state_dict(self.lr_scheduler_state_dict)

        # Parse loss from training configuration
        # Loss can be changed with a custom one here!
        self._set_loss_from_config()

        # IMPORTANT: model, optimizer, and scheduler need to be distributed from here on

        distribute_kwargs = self.get_default_distributed_kwargs()

        # Distributed model, optimizer, and scheduler
        (self.model, self.optimizer, self.lr_scheduler) = self.strategy.distributed(
            self.model, self.optimizer, self.lr_scheduler, **distribute_kwargs
        )

    def save_checkpoint(
        self,
        name: str,
        best_validation_metric: torch.Tensor | None = None,
        checkpoints_root: str | Path | None = None,
        force: bool = False,
    ) -> str | None:
        """Save training checkpoint.

        Args:
            name (str): name of the checkpoint directory.
            best_validation_metric (torch.Tensor | None): best validation metric throughout
                training so far (if available). Usually this is the validation loss.
            checkpoints_root (str | None): path for root checkpoints dir. If None, uses
                ``self.checkpoints_location`` as base.
            force (bool): force checkpointign now.

        Returns:
            path to the checkpoint file or ``None`` when the checkpoint is not created.
        """
        # Determine whether a checkpoint should be created
        should_checkpoint = self.strategy.is_main_worker and (
            force
            or self.checkpoint_every
            and (self.current_epoch + 1) % self.checkpoint_every == 0
        )

        ckpt_dir = Path(checkpoints_root or self.checkpoints_location) / name
        py_logger.info(f"Saving checkpoint at {ckpt_dir.resolve()}? {should_checkpoint}")

        if not should_checkpoint:
            # Do nothing and return
            return

        ckpt_dir = Path(checkpoints_root or self.checkpoints_location) / name
        ckpt_dir.mkdir(parents=True, exist_ok=True)

        # Save state (epoch, loss, optimizer, scheduler)
        state = {
            "epoch": self.current_epoch,
            # This could store the best validation loss
            "best_validation_metric": (
                best_validation_metric.item() if best_validation_metric is not None else None
            ),
            "optimizer_state_dict": self.optimizer.state_dict(),
            "lr_scheduler_state_dict": (
                self.lr_scheduler.state_dict() if self.lr_scheduler is not None else None
            ),
            "torch_rng_state": self.torch_rng.get_state(),
            "random_seed": self.random_seed,
        }
        state_path = ckpt_dir / "state.pt"
        torch.save(state, state_path)

        # Save PyTorch model separately
        model_path = ckpt_dir / "model.pt"
        # TODO: check that the state dict is stripped from any distributed info
        torch.save(self.model.state_dict(), model_path)

        # Save Pydantic config as YAML
        config_path = ckpt_dir / "config.yaml"
        with config_path.open("w") as f:
            yaml.safe_dump(self.config.model_dump(), f)

        # Log each file with an appropriate identifier
        self.log(str(state_path), f"{name}_state", kind="artifact")
        self.log(str(model_path), f"{name}_model", kind="artifact")
        self.log(str(config_path), f"{name}_config", kind="artifact")

        assert state_path.exists()
        assert model_path.exists()
        assert config_path.exists()

        py_logger.info(f"Saved checkpoint at {ckpt_dir.resolve()}")

        return str(ckpt_dir)

    def load_checkpoint(self) -> None:
        """Reload training state from checkpoint."""
        if not self.from_checkpoint:
            # A checkpoint path was NOT provided
            return

        # A checkpoint path was provided
        py_logger.info(f"Loading from existing checkpoint at {self.from_checkpoint}")

        if not isinstance(self.strategy, RayTorchDistributedStrategy):
            # Not using Ray, falling back to simple checkpoint reload
            py_logger.debug("Loading from existing checkpoint without using Ray")
            self._load_checkpoint(checkpoint_dir=self.from_checkpoint)
            return

        # A Ray checkpoint directory was passed to the trainer -- assuming to be inside a trial
        checkpoint = ray.train.get_checkpoint()
        if not checkpoint:
            py_logger.warning(
                "A checkpoint path was passed, but Ray could not find a valid "
                "checkpoint directory. Skipping loading from checkpoint."
            )
            return
        with checkpoint.as_directory() as checkpoint_dir:
            py_logger.debug("Loading from existing Ray checkpoint")
            self._load_checkpoint(checkpoint_dir=checkpoint_dir)

    def _load_checkpoint(self, checkpoint_dir: str | Path) -> None:
        """Load checkpoint from path."""
        checkpoint_dir = Path(checkpoint_dir)
        state = torch.load(checkpoint_dir / "state.pt")

        # Override initial training state
        self.model_state_dict = torch.load(checkpoint_dir / "model.pt")
        self.optimizer_state_dict = state["optimizer_state_dict"]
        self.lr_scheduler_state_dict = state["lr_scheduler_state_dict"]
        self.torch_rng_state = state["torch_rng_state"]
        # Direct overrides (don't require further attention)
        self.random_seed = state["random_seed"]
        self.current_epoch = state["epoch"] + 1  # Start from next epoch
        if state["best_validation_metric"]:
            self.best_validation_metric = state["best_validation_metric"]

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> None:
        """
        Create train, validation and test dataloaders using the
        configuration provided in the Trainer constructor.
        Generally a user-defined method.

        Args:
            train_dataset (Dataset): training dataset object.
            validation_dataset (Dataset | None): validation dataset object.
                Default None.
            test_dataset (Dataset | None): test dataset object.
                Default None.
        """

        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################
        self.train_dataloader = self.strategy.create_dataloader(
            dataset=train_dataset,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers_dataloader,
            pin_memory=self.config.pin_gpu_memory,
            generator=self.torch_rng,
            shuffle=self.config.shuffle_train,
        )
        if validation_dataset is not None:
            self.validation_dataloader = self.strategy.create_dataloader(
                dataset=validation_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers_dataloader,
                pin_memory=self.config.pin_gpu_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_validation,
            )
        if test_dataset is not None:
            self.test_dataloader = self.strategy.create_dataloader(
                dataset=test_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers_dataloader,
                pin_memory=self.config.pin_gpu_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_test,
            )

    def _setup_metrics(self) -> None:
        """Move metrics to current device."""
        for m_name, metric in self.metrics.items():
            self.metrics[m_name] = metric.to(self.device)

    @monitor_exec
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        """Prepares distributed environment and data structures
        for the actual training.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Dataset | None, optional): validation
                dataset. Defaults to None.
            test_dataset (Dataset | None, optional): test dataset.
                Defaults to None.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        if isinstance(self.strategy, RayTorchDistributedStrategy):
            # Execute with Ray
            return self._execute_with_ray(
                train_dataset=train_dataset,
                validation_dataset=validation_dataset,
                test_dataset=test_dataset,
            )

        # Execute without ray
        if self.ray_scaling_config:
            py_logger.warning(
                "Ray scaling config was passed, but it's ignored as Ray is not used"
            )
        if self.ray_run_config:
            py_logger.warning("Ray run config was passed, but it's ignored as Ray is not used")
        if self.ray_tune_config:
            py_logger.warning(
                "Ray tune config was passed, but it's ignored as Ray is not used"
            )
        if self.ray_search_space:
            py_logger.warning(
                "Ray search space was passed, but it's ignored as Ray is not used"
            )
        if self.ray_horovod_config:
            py_logger.warning(
                "Ray horovod config was passed, but it's ignored as Ray is not used"
            )
        if self.ray_torch_config:
            py_logger.warning(
                "Ray torch config was passed, but it's ignored as Ray is not used"
            )
        if self.ray_data_config:
            py_logger.warning(
                "Ray dataset config was passed, but it's ignored as Ray is not used"
            )

        self._run_worker(
            config={},
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )
        return train_dataset, validation_dataset, test_dataset, None

    def _execute_with_ray(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        """Launch training and, optionally, hyperarameter tuning with Ray"""

        train_with_data = ray.tune.with_parameters(
            self._run_worker,
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )

        if self.ray_run_config:
            # Create Ray checkpoints dir if it does not exist yet
            ckpt_dir = Path(self.ray_run_config.storage_path)
            ckpt_dir.mkdir(parents=True, exist_ok=True)

        if isinstance(self.strategy, RayHorovodStrategy):
            # Using Horovod with Ray
            from ray.train.horovod import HorovodTrainer

            if self.ray_torch_config:
                py_logger.warning(
                    "Ray torch config was passed, but it's ignored as "
                    f"{self.strategy.__class__.__name__} strategy is used"
                )

            if self.from_checkpoint:
                # Create trainer from checkpoint
                if HorovodTrainer.can_restore(to_uri(self.from_checkpoint)):
                    trainer = HorovodTrainer.restore(
                        path=to_uri(self.from_checkpoint),
                        train_loop_per_worker=train_with_data,
                        train_loop_config=None,
                    )
                else:
                    # Ray is unable to restore the checkpoint implicitly, but it's passing
                    # it to the trial
                    trainer = HorovodTrainer(
                        train_loop_per_worker=train_with_data,
                        train_loop_config=None,
                        horovod_config=self.ray_horovod_config,
                        scaling_config=self.ray_scaling_config,
                        run_config=self.ray_run_config,
                        dataset_config=self.ray_data_config,
                        resume_from_checkpoint=Checkpoint(to_uri(self.from_checkpoint)),
                    )
            else:
                # Create trainer without checkpoint
                trainer = HorovodTrainer(
                    train_loop_per_worker=train_with_data,
                    train_loop_config=None,
                    horovod_config=self.ray_horovod_config,
                    scaling_config=self.ray_scaling_config,
                    run_config=self.ray_run_config,
                    dataset_config=self.ray_data_config,
                )
        else:
            # Using DDP or DeepSpeed with Ray
            if self.ray_horovod_config:
                py_logger.warning(
                    "Ray horovod config was passed, but it's ignored as "
                    f"{self.strategy.__class__.__name__} strategy is used"
                )
            if self.from_checkpoint:
                # Create trainer from checkpoint
                if RayTorchTrainer.can_restore(to_uri(self.from_checkpoint)):
                    trainer = RayTorchTrainer.restore(
                        path=to_uri(self.from_checkpoint),
                        train_loop_per_worker=train_with_data,
                        train_loop_config=None,
                    )
                else:
                    # Ray is unable to restore the checkpoint implicitly, but it's passing
                    # it to the trial
                    trainer = RayTorchTrainer(
                        train_loop_per_worker=train_with_data,
                        train_loop_config=None,
                        scaling_config=self.ray_scaling_config,
                        run_config=self.ray_run_config,
                        torch_config=self.ray_torch_config,
                        dataset_config=self.ray_data_config,
                        resume_from_checkpoint=Checkpoint(to_uri(self.from_checkpoint)),
                    )
            else:
                # Create trainer without checkpoint
                trainer = RayTorchTrainer(
                    train_loop_per_worker=train_with_data,
                    train_loop_config=None,
                    scaling_config=self.ray_scaling_config,
                    run_config=self.ray_run_config,
                    torch_config=self.ray_torch_config,
                    dataset_config=self.ray_data_config,
                )

        # Wrap the trainer into a Tuner
        param_space = {"train_loop_config": search_space(self.ray_search_space)}
        tuner = ray.tune.Tuner(
            trainable=trainer,
            param_space=param_space,
            tune_config=self.ray_tune_config,
        )

        if self.time_ray:
            self.tune_result_grid = self._time_and_log(
                lambda: tuner.fit(), "ray_fit_time_s", step=0
            )
        else:
            self.tune_result_grid = tuner.fit()

        return train_dataset, validation_dataset, test_dataset, None

    def _run_worker(
        self,
        config: Dict,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> None:
        self.load_checkpoint()

        self._override_config(config)

        self._set_seed()
        self._init_distributed_strategy()
        self._setup_metrics()

        if self.logger:
            py_logger.debug(f"Using logger: {self.logger.__class__.__name__}")
            self.logger.create_logger_context(rank=self.strategy.global_rank())
            py_logger.debug("...the logger has been initialized")
            hparams = self.config.model_dump()
            hparams["distributed_strategy"] = self.strategy.__class__.__name__
            self.logger.save_hyperparameters(hparams)

        self.create_dataloaders(
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )
        self.create_model_loss_optimizer()

        self.train()

        if self.logger:
            self.logger.destroy_logger_context()

        self.strategy.clean_up()

    def _set_seed(self) -> None:
        py_logger.debug(f"Using random seed: {self.random_seed}")
        self.torch_rng = set_seed(self.random_seed)

        if self.torch_rng_state is not None:
            # Resume state from checkpoint
            py_logger.debug("Resuming torch PRNG state from checkpoint")
            self.torch_rng.set_state(self.torch_rng_state)

    def _override_config(self, config: Dict) -> None:
        """Overrid self.config with a sample from the search space from the Ray tuner."""
        self.config = self.config.model_copy(update=config)
        py_logger.debug("Overridden self.config with trial config (if given)")

    def _set_epoch_dataloaders(self, epoch: int) -> None:
        """Sets epoch in the distributed sampler of a dataloader when using it."""
        if self.strategy.is_distributed:
            self.train_dataloader.sampler.set_epoch(epoch)
            if self.validation_dataloader is not None:
                self.validation_dataloader.sampler.set_epoch(epoch)
            if self.test_dataloader is not None:
                self.test_dataloader.sampler.set_epoch(epoch)

    def set_epoch(self) -> None:
        """Set current epoch at the beginning of training."""
        if self.profiler is not None and self.current_epoch > 0:
            # We don't want to start stepping until after the first epoch
            self.profiler.step()
        if self.lr_scheduler:
            self.lr_scheduler.step()
        self._set_epoch_dataloaders(self.current_epoch)

    def log(
        self,
        item: Any | List[Any],
        identifier: str | List[str],
        kind: str = "metric",
        step: int | None = None,
        batch_idx: int | None = None,
        **kwargs,
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Any | List[Any]): element to be logged (e.g., metric).
            identifier (str | List[str]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (int | None, optional): logging step. Defaults to None.
            batch_idx (int | None, optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """
        if self.logger:
            self.logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs,
            )

    def ray_report(
        self,
        metrics: Dict[str, float],
        checkpoint_file: str | Path | None = None,
        checkpoint_dir: str | Path | None = None,
        checkpoint_data: Any | None = None,
    ) -> None:
        """Report a dictionary of metrics and optionally a checkpoint to Ray, only when using
        Ray distributed strategies. The checkpoint could be in the form of a Python object
        (passed as ``checkpoint_data``), the path to a single file (passed as
        ``checkpoint_file``), or the path to an existing checkpoint directory (passed as
        ``checkpoint_dir``).

        Args:
            metrics (Dict[str, float]): metrics to be reported.
            checkpoint_file (str | Path | None, optional): path to the checkpoint file.
                Defaults to None.
            checkpoint_dir (str | Path | None, optional): path to the checkpoint directory.
                Defaults to None.
            checkpoint_data (Any | None, optional):object to serialize as a checkpoint.
                Defaults to None.
        """
        if not isinstance(self.strategy, RayTorchDistributedStrategy):
            # Ray is not used, thus do nothing
            return

        if checkpoint_file:
            # A checkpoint is given as a file
            with tempfile.TemporaryDirectory() as tmp_dir:
                import shutil

                shutil.copy(checkpoint_file, tmp_dir)
                checkpoint = ray.train.Checkpoint.from_directory(tmp_dir)
                ray.train.report(metrics, checkpoint=checkpoint)

        elif checkpoint_data:
            # A checkpoint is given as a python object which needs to be serialized
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp_dir = Path(tmp_dir)
                ckpt_file = tmp_dir / "ckpt.pt"
                torch.save(checkpoint_data, ckpt_file)
                checkpoint = ray.train.Checkpoint.from_directory(tmp_dir)
                ray.train.report(metrics, checkpoint=checkpoint)

        elif checkpoint_dir:
            # A checkpoint is given as a directory
            checkpoint = ray.train.Checkpoint.from_directory(checkpoint_dir)
            ray.train.report(metrics, checkpoint=checkpoint)

        else:
            # No checkpoint is given: only report metrics
            ray.train.report(metrics)

    def compute_metrics(
        self,
        true: Batch,
        pred: Batch,
        logger_step: int,
        batch_idx: int | None,
        stage: str = "train",
    ) -> Dict[str, Any]:
        """Compute and log metrics.

        Args:
            metrics (Dict[str, Callable]): metrics dict. Can be
                ``self.train_metrics`` or ``self.validation_metrics``.
            true (Batch): true values.
            pred (Batch): predicted values.
            logger_step (int): global step to pass to the logger.
            stage (str): 'train', 'validation'...

        Returns:
            Dict[str, Any]: metric values.
        """
        m_values = {}
        for m_name, metric in self.metrics.items():
            # metric = metric.to(self.device)
            m_val = metric(pred, true).detach().cpu().numpy()
            self.log(
                item=m_val,
                identifier=f"{stage}_{m_name}",
                kind="metric",
                step=logger_step,
                batch_idx=batch_idx,
            )
            m_values[m_name] = m_val
        return m_values

    @profile_torch_trainer
    @measure_gpu_utilization
    def train(self) -> None:
        """Trains a machine learning model.
        Main training loop/logic.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Dataset): validation dataset.
            test_dataset (Dataset): test dataset.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        epoch_time_logger: EpochTimeTracker | None = None
        if self.strategy.is_main_worker and self.strategy.is_distributed:
            if "SLURM_NNODES" not in os.environ:
                raise EnvironmentError(
                    "'SLURM_NNODES' is not present in 'os.environ', but is required"
                    " when running distributed training!"
                )
            num_nodes = int(os.environ["SLURM_NNODES"])
            epoch_time_output_dir = Path(f"scalability-metrics/{self.run_id}/epoch-time")
            epoch_time_file_name = f"epochtime_{self.strategy.name}_{num_nodes}N.csv"
            epoch_time_output_path = epoch_time_output_dir / epoch_time_file_name

            epoch_time_logger = EpochTimeTracker(
                strategy_name=self.strategy.name,
                save_path=epoch_time_output_path,
                num_nodes=num_nodes,
                should_log=self.measure_epoch_time,
            )

        progress_bar = tqdm(
            range(self.current_epoch, self.epochs),
            desc="Epochs",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
        )

        for self.current_epoch in progress_bar:
            epoch_start_time = default_timer()
            progress_bar.set_description(f"Epoch {self.current_epoch + 1}/{self.epochs}")

            self.set_epoch()
            self.train_epoch()
            val_metric = self.validation_epoch()

            # Periodic checkpointing
            periodic_ckpt_path = self.save_checkpoint(name=f"epoch_{self.current_epoch}")

            # Checkpointing current best model
            best_ckpt_path = None
            worker_val_metrics = self.strategy.gather(val_metric, dst_rank=0)

            if self.strategy.is_main_worker:
                avg_metric = torch.mean(torch.stack(worker_val_metrics)).detach().cpu()
                if avg_metric < self.best_validation_metric:
                    best_ckpt_path = self.save_checkpoint(
                        name="best_model",
                        best_validation_metric=avg_metric,
                        force=True,
                    )
                    self.best_validation_metric = avg_metric

            # Report validation metrics to Ray (useful for tuning!)
            metric_name = _get_tuning_metric_name(self.ray_tune_config)
            if metric_name is None:
                raise ValueError("Could not find a metric in the TuneConfig")

            if self.time_ray:
                # time and log the ray_report call
                self._time_and_log(
                    lambda: self.ray_report(
                        metrics={metric_name: val_metric.item()},
                        checkpoint_dir=best_ckpt_path or periodic_ckpt_path,
                    ),
                    "ray_report_time_s_per_epoch",
                    step=self.current_epoch,
                )
            else:
                self.ray_report(
                    metrics={metric_name: val_metric.item()},
                    checkpoint_dir=best_ckpt_path or periodic_ckpt_path,
                )

            if self.test_every and (self.current_epoch + 1) % self.test_every == 0:
                self.test_epoch()

            if self.strategy.is_main_worker and self.strategy.is_distributed:
                assert epoch_time_logger is not None
                epoch_time = default_timer() - epoch_start_time
                epoch_time_logger.add_epoch_time(self.current_epoch + 1, epoch_time)

    def train_epoch(self) -> torch.Tensor:
        """Perform a complete sweep over the training dataset, completing an
        epoch of training.

        Args:
            epoch (int): current epoch number, from 0 to ``self.epochs - 1``.

        Returns:
            Loss: average training loss for the current epoch.
        """
        self.model.train()
        train_loss_sum = 0.0
        train_metrics_sum = defaultdict(float)
        batch_counter = 0

        progress_bar = tqdm(
            enumerate(self.train_dataloader),
            total=len(self.train_dataloader) // self.strategy.global_world_size(),
            desc="Train batches",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
            leave=False,  # Set this to true to see how many batches were used
        )

        for batch_idx, train_batch in progress_bar:
            loss, metrics = self.train_step(batch=train_batch, batch_idx=batch_idx)
            train_loss_sum += loss
            batch_counter += 1
            for name, val in metrics.items():
                train_metrics_sum[name] += val

            # Important: update counter
            self.train_glob_step += 1

        # Aggregate and log losses
        avg_loss = train_loss_sum / batch_counter
        self.log(
            item=avg_loss.item(),
            identifier="train_loss_epoch",
            kind="metric",
            step=self.train_glob_step,
        )
        # Aggregate and log metrics
        for m_name, m_val in train_metrics_sum.items():
            self.log(
                item=m_val / batch_counter,
                identifier="train_" + m_name + "_epoch",
                kind="metric",
                step=self.train_glob_step,
            )

        return avg_loss

    def train_step(self, batch: Batch, batch_idx: int) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        training dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)

        self.optimizer.zero_grad()
        pred_y = self.model(x)
        loss = self.loss(pred_y, y)
        loss.backward()
        self.optimizer.step()

        # Log metrics
        self.log(
            item=loss.item(),
            identifier="train_loss",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.train_glob_step,
            batch_idx=batch_idx,
            stage="train",
        )
        return loss, metrics

    def validation_epoch(self) -> torch.Tensor:
        """Perform a complete sweep over the validation dataset, completing an
        epoch of validation.

        Returns:
            Loss | None: average validation loss for the current epoch if
                self.validation_dataloader is not None
        """
        if self.validation_dataloader is None:
            return

        progress_bar = tqdm(
            enumerate(self.validation_dataloader),
            total=len(self.validation_dataloader) // self.strategy.global_world_size(),
            desc="Validation batches",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
            leave=False,  # Set this to true to see how many batches were used
        )

        self.model.eval()
        validation_loss_sum = 0.0
        validation_metrics_sum = defaultdict(float)
        batch_counter = 0
        for batch_idx, val_batch in progress_bar:
            loss, metrics = self.validation_step(batch=val_batch, batch_idx=batch_idx)
            validation_loss_sum += loss
            batch_counter += 1
            for name, val in metrics.items():
                validation_metrics_sum[name] += val

            # Important: update counter
            self.validation_glob_step += 1

        # Aggregate and log losses
        avg_loss = validation_loss_sum / batch_counter
        self.log(
            item=avg_loss.item(),
            identifier="validation_loss_epoch",
            kind="metric",
            step=self.validation_glob_step,
        )
        # Aggregate and log metrics
        for m_name, m_val in validation_metrics_sum.items():
            self.log(
                item=m_val / batch_counter,
                identifier="validation_" + m_name + "_epoch",
                kind="metric",
                step=self.validation_glob_step,
            )

        return avg_loss

    def validation_step(
        self, batch: Batch, batch_idx: int
    ) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        validation dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)
        with torch.no_grad():
            pred_y = self.model(x)
            loss: torch.Tensor = self.loss(pred_y, y)
        self.log(
            item=loss.item(),
            identifier="validation_loss",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.validation_glob_step,
            batch_idx=batch_idx,
            stage="validation",
        )
        return loss, metrics

    def test_epoch(self) -> torch.Tensor:
        """Perform a complete sweep over the test dataset, completing an
        epoch of test.

        Returns:
            Loss: average test loss for the current epoch.
        """
        raise NotImplementedError()

    def test_step(self, batch: Batch, batch_idx: int) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single predictions step using a batch sampled from the
        test dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        raise NotImplementedError()


class TorchLightningTrainer(Trainer):
    """Generic trainer for torch Lightning workflows.

    Args:
        config (Dict | str): `Lightning configuration`_
            which can be the path to a file or a Python dictionary.
        mlflow_saved_model (str, optional): name of the model created in
            MLFlow. Defaults to 'my_model'.

    .. _Lightning configuration:
        https://pytorch-lightning.readthedocs.io/en/1.6.5/common/lightning_cli.html
    """

    def __init__(self, config: Dict | str, mlflow_saved_model: str = "my_model"):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config
        self.mlflow_saved_model = mlflow_saved_model

    @monitor_exec
    def execute(self) -> Any:
        import lightning as L
        from lightning.pytorch.cli import LightningCLI

        from .mlflow import init_lightning_mlflow, teardown_lightning_mlflow

        init_lightning_mlflow(
            self.conf, tmp_dir="/tmp", registered_model_name=self.mlflow_saved_model
        )
        old_argv = sys.argv
        sys.argv = ["some_script_placeholder.py"]
        cli = LightningCLI(
            args=self.conf,
            model_class=L.LightningModule,
            datamodule_class=L.LightningDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv
        cli.trainer.fit(cli.model, datamodule=cli.datamodule)
        teardown_lightning_mlflow()


def _distributed_dataloader(dataloader: DataLoader, gwsize, grank):
    """Makes a Dataloader distributed."""
    sampler = DistributedSampler(
        dataloader.dataset, num_replicas=gwsize, rank=grank, shuffle=True
    )
    # Recreate dataloader, with updated sampler
    return DataLoader(
        dataloader.dataset,
        batch_size=dataloader.batch_size,
        sampler=sampler,
        num_workers=dataloader.num_workers,
        collate_fn=dataloader.collate_fn,
        pin_memory=dataloader.pin_memory,
        drop_last=dataloader.drop_last,
        timeout=dataloader.timeout,
        worker_init_fn=seed_worker,  # dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
        generator=dataloader.generator,
        prefetch_factor=dataloader.prefetch_factor,
        persistent_workers=dataloader.persistent_workers,
        pin_memory_device=dataloader.pin_memory_device,
    )


def distributed(func):
    """The decorated function must have a standard signature.
    Its first arguments must be:
    model, train_dataloader, validation_dataloader, device (in this order).

    Additional args or kwargs are allowed consistently with the signature
    of the decorated function.
    """

    def dist_train(
        model, train_dataloader, validation_dataloader=None, device="cpu", *args, **kwargs
    ):
        if torch.cuda.is_available():
            dist.init_process_group(backend="nccl")

        if torch.cuda.is_available():
            lwsize = torch.cuda.device_count()  # local world size - per node
            gwsize = dist.get_world_size()  # global world size - per run
            grank = dist.get_rank()  # global rank - assign per run
            lrank = dist.get_rank() % lwsize  # local rank - assign per node
        else:
            gwsize = 1
            grank = 0
            lrank = 0

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu", lrank)
        if torch.cuda.is_available():
            torch.cuda.set_device(lrank)

        model = model.to(device)
        model = DDP(model, device_ids=[device], output_device=device)

        train_dataloader = _distributed_dataloader(train_dataloader, gwsize, grank)
        if validation_dataloader is not None:
            validation_dataloader = _distributed_dataloader(
                validation_dataloader, gwsize, grank
            )

        try:
            func(model, train_dataloader, validation_dataloader, device, *args, **kwargs)
        finally:
            if torch.cuda.is_available():
                dist.barrier()
                dist.destroy_process_group()

    return dist_train



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from unittest.mock import MagicMock, patch

import pytest

from itwinai.distributed import get_adaptive_ray_scaling_config, ray_cluster_is_running
from itwinai.torch.trainer import TorchTrainer


@pytest.mark.parametrize(
    "strategy_fixture",
    [
        pytest.param(None),  # NonDistributedStrategy
        pytest.param(
            "ddp_strategy",
            marks=[pytest.mark.torch_dist, pytest.mark.hpc],
        ),
        pytest.param(
            "deepspeed_strategy",
            marks=[pytest.mark.deepspeed_dist, pytest.mark.hpc],
        ),
        pytest.param(
            "horovod_strategy",
            marks=[pytest.mark.horovod_dist, pytest.mark.hpc],
        ),
    ],
)
def test_distributed_trainer_mnist(
    mnist_datasets, request, strategy_fixture, named_temp_dir, mnist_net
):
    """Test TorchTrainer on MNIST with different distributed strategies."""
    training_config = dict(optimizer="sgd", loss="nllloss")
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=2,
        strategy=None,
        checkpoint_every=1,
        checkpoints_location=named_temp_dir / "my_checkpoints",
    )
    if strategy_fixture:
        # Override when the strategy is supposed to be distributed
        trainer.strategy = request.getfixturevalue(strategy_fixture)

    train_set, val_set = mnist_datasets

    # Mock strategy cleanup -- IMPORTANT, otherwise the trainer will mess up with the strategy
    # fixture
    with patch.object(
        trainer.strategy, "clean_up", new=MagicMock(name="clean_up")
    ) as mock_cleanup:
        trainer.execute(train_set, val_set)

        # Check that the torch trainer is cleaning up the strategy
        mock_cleanup.assert_called_once()

    # Restore training from checkpoint
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=1,
        strategy=None,
        checkpoint_every=1,
        from_checkpoint=named_temp_dir / "my_checkpoints/best_model",
        checkpoints_location=named_temp_dir / "my_checkpoints_new_model",
    )
    # Mock strategy cleanup -- IMPORTANT, otherwise the trainer will mess up with the strategy
    # fixture
    with patch.object(
        trainer.strategy, "clean_up", new=MagicMock(name="clean_up")
    ) as mock_cleanup:
        trainer.execute(train_set, val_set)

        # Check that the torch trainer is cleaning up the strategy
        mock_cleanup.assert_called_once()


@pytest.mark.hpc
@pytest.mark.ray_dist
@pytest.mark.parametrize(
    "strategy_name",
    [
        pytest.param("ddp"),
        pytest.param("deepspeed"),
        pytest.param("horovod"),
    ],
)
def test_distributed_trainer_mnist_ray(
    mnist_datasets, strategy_name, shared_tmp_path, mnist_net
):
    """Test TorchTrainer on MNIST with different distributed strategies using Ray."""
    from ray.train import RunConfig

    assert ray_cluster_is_running(), "Ray cluster not detected. Aborting tests"

    ray_run_config = RunConfig(storage_path=shared_tmp_path / "ray_checkpoints")

    ckpt_path = shared_tmp_path / "my_checkpoints" / strategy_name
    training_config = dict(optimizer="sgd", loss="nllloss")
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=2,
        strategy=strategy_name,
        ray_scaling_config=get_adaptive_ray_scaling_config(),
        ray_run_config=ray_run_config,
        checkpoint_every=1,
        checkpoints_location=ckpt_path,
    )

    train_set, val_set = mnist_datasets

    # Train
    # TODO: prevent strategy cleanup?
    trainer.execute(train_set, val_set)

    # Restore training from checkpoint
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=1,
        strategy=strategy_name,
        ray_scaling_config=get_adaptive_ray_scaling_config(),
        checkpoint_every=1,
        from_checkpoint=ckpt_path / "best_model",
        checkpoints_location=ckpt_path,
    )
    # Resume training
    # TODO: prevent strategy cleanup?
    trainer.execute(train_set, val_set)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import os
import subprocess
from typing import Callable

import pytest

FNAMES = [
    "pipeline.yaml",
    "startscript",
]


@pytest.fixture
def check_folder_structure() -> Callable:
    """Verify that the use case folder complies with some predefined structure."""

    def _check_structure(root: str):
        for fname in FNAMES:
            fpath = os.path.join(root, fname)
            assert os.path.isfile(fpath), f"'{fname}' is missing in '{fpath}'"

    return _check_structure


@pytest.fixture
def install_requirements() -> Callable:
    """Install requirements.txt, if present in root folder."""

    def _install_reqs(root: str, env_prefix: str):
        req_path = os.path.join(root, "requirements.txt")
        if os.path.isfile(req_path):
            cmd = f"{env_prefix}/bin/pip install --no-cache-dir -r {req_path}"
            subprocess.run(cmd.split(), check=True)

    return _install_reqs



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


import copy
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import LinearLR
from torch.utils.data import Dataset

from itwinai.torch.trainer import TorchTrainer


# Sample training configuration mock
def get_mock_config():
    return {
        "optim_lr": 0.001,
        "optimizer": "adam",
        "batch_size": 5,
        "loss": "mse",
        "lr_scheduler": "linear",
    }


class DummyDataset(Dataset):
    def __init__(self, size=100):
        self.size = size
        self.data = torch.randn(size, 10)
        self.targets = torch.randint(0, 2, (size,))

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.data[idx].type(torch.float), self.targets[idx].type(torch.float)


def assert_equal_optimizer_states(state1, state2):
    """Compares two optimizer state dictionaries."""

    """Compares two optimizer state dictionaries with detailed assertions."""
    assert state1.keys() == state2.keys(), "Optimizer state keys mismatch"

    # Compare param_groups excluding 'params' because indexes might differ
    for key in state1["param_groups"][0]:
        if key != "params":
            assert state1["param_groups"][0][key] == state2["param_groups"][0][key], (
                f"Mismatch in param_groups key: {key}"
            )

    # Ensure both have the same parameter keys
    assert state1["state"].keys() == state2["state"].keys(), (
        "Parameter keys in optimizer state do not match"
    )

    # Compare each parameter state
    for param_id in state1["state"]:
        assert param_id in state2["state"], (
            f"Parameter {param_id} missing in second optimizer state"
        )

        for subkey in state1["state"][param_id]:
            tensor1 = state1["state"][param_id][subkey]
            tensor2 = state2["state"][param_id].get(subkey)

            assert tensor2 is not None, (
                f"Missing key '{subkey}' for param {param_id} in second state"
            )

            assert torch.allclose(tensor1, tensor2, atol=1e-6), (
                f"Mismatch in {subkey} for param {param_id}"
            )


def to_cpu_recursive(state_dict):
    from collections import OrderedDict

    """Recursively moves all tensors in a state_dict to CPU while preserving OrderedDict
    structure.
    """
    if isinstance(state_dict, OrderedDict):
        return OrderedDict((k, to_cpu_recursive(v)) for k, v in state_dict.items())
    elif isinstance(state_dict, dict):  # Handle nested dicts
        return {k: to_cpu_recursive(v) for k, v in state_dict.items()}
    elif isinstance(state_dict, torch.Tensor):  # Move tensors to CPU
        return state_dict.to("cpu")
    return state_dict  # Return unchanged for non-tensor values


def equal_models(model1: torch.nn.Module, model2: torch.nn.Module) -> bool:
    # Compare the state dictionaries
    state_dict1 = to_cpu_recursive(model1.state_dict())
    state_dict2 = to_cpu_recursive(model2.state_dict())

    if state_dict1.keys() != state_dict2.keys():
        return False

    for key in state_dict1:
        if not torch.equal(state_dict1[key], state_dict2[key]):
            print(f"Mismatch in parameter: {key}")
            return False

    return True


def test_inexistent_checkpoint():
    # Test unexisting checkpoint directory
    with pytest.raises(RuntimeError) as err:
        TorchTrainer(
            config=get_mock_config(),
            epochs=2,
            model=None,
            strategy="ddp",
            from_checkpoint="/my/checkpoint/dir",
            checkpoint_every=1,
        )
        assert "checkpoint is not found" in err


@pytest.mark.parametrize(
    "strategy_fixture",
    [
        pytest.param(None),
        pytest.param("ddp_strategy", marks=[pytest.mark.hpc, pytest.mark.torch_dist]),
        pytest.param(
            "deepspeed_strategy", marks=[pytest.mark.hpc, pytest.mark.deepspeed_dist]
        ),
        pytest.param("horovod_strategy", marks=[pytest.mark.hpc, pytest.mark.horovod_dist]),
    ],
)
def test_checkpoint_loading(strategy_fixture: str | None, named_temp_dir: Path, request):
    """Test loading of checkpoints that were generated by some distributed strategy.
    This is ignoring Ray strategies.
    """
    model = torch.nn.Linear(10, 1)
    untrained_model = copy.deepcopy(model)
    ckpt_path = named_temp_dir / "checkpoint"
    ckpt_path.mkdir(parents=True, exist_ok=True)

    assert ckpt_path.exists(), f"Failed to create checkpoints path at {ckpt_path.resolve()}"

    train_dataset = DummyDataset()
    validation_dataset = DummyDataset()

    # Create cehckpoint throigh the TorchTrainer execution
    trainer = TorchTrainer(
        config=get_mock_config(),
        epochs=2,
        model=model,
        strategy=None,
        checkpoint_every=1,
        checkpoints_location=ckpt_path,
        random_seed=42,
    )

    if strategy_fixture:
        # Patch trainer's strategy with strategy fixture
        trainer.strategy = request.getfixturevalue(strategy_fixture)

    # Mock some methods to skip training
    best_val_loss = 0.5
    trainer.validation_epoch = MagicMock()
    # Set list of return values for successive calls -- simulate reduction in val loss
    trainer.validation_epoch.side_effect = [torch.tensor(0.9), torch.tensor(best_val_loss)]

    # Run training: generate checkpoints
    # Mock strategy cleanup -- IMPORTANT, otherwise the trainer will mess up with the strategy
    # fixture
    with patch.object(
        trainer.strategy, "clean_up", new=MagicMock(name="clean_up")
    ) as mock_cleanup:
        trainer.execute(train_dataset=train_dataset, validation_dataset=validation_dataset)

        # Check that the torch trainer is cleaning up the strategy
        mock_cleanup.assert_called_once()

    # Check that the model was actually trained
    assert not equal_models(trainer.model, untrained_model)

    best_ckpt_path = ckpt_path / "best_model"
    assert best_ckpt_path.exists(), (
        f"Could not find best checkpoint path at {best_ckpt_path.resolve()}. "
        f"Checkpoints location contains: {list(ckpt_path.glob('**/**'))}"
    )

    # Create TorchTrainer instance with checkpoint
    model = torch.nn.Linear(10, 1)
    target_ckpt = best_ckpt_path
    trainer = TorchTrainer(
        config=get_mock_config(),
        epochs=2,
        model=model,
        strategy=None,
        from_checkpoint=target_ckpt,
    )

    trainer.create_dataloaders = MagicMock()
    trainer.train = MagicMock()
    # Run training: force checkpoint to be loaded
    # Mock strategy cleanup -- IMPORTANT, otherwise the trainer will mess up with the strategy
    # fixture
    with patch.object(
        trainer.strategy, "clean_up", new=MagicMock(name="clean_up")
    ) as mock_cleanup:
        trainer.execute(train_dataset=MagicMock(), validation_dataset=MagicMock())

        # Check that the torch trainer is cleaning up the strategy
        mock_cleanup.assert_called_once()

    # Manually simulate the loading of checkpoint
    state = torch.load(target_ckpt / "state.pt", weights_only=True)
    model_state_dict = torch.load(target_ckpt / "model.pt", weights_only=True)
    optimizer_state_dict = state["optimizer_state_dict"]
    lr_scheduler_state_dict = state["lr_scheduler_state_dict"]
    torch_rng_state = state["torch_rng_state"]
    random_seed = state["random_seed"]
    epoch = state["epoch"]
    best_validation_loss = None
    if state["best_validation_metric"]:
        best_validation_loss = state["best_validation_metric"]
    loaded_model = torch.nn.Linear(10, 1)
    loaded_model.load_state_dict(model_state_dict, strict=False)
    loaded_optim = Adam(loaded_model.parameters())
    loaded_lr_scheduler = LinearLR(loaded_optim)
    # Remember to load optim state dict after the scheduler has already been instantiated
    loaded_optim.load_state_dict(optimizer_state_dict)
    loaded_lr_scheduler.load_state_dict(lr_scheduler_state_dict)

    # Best validation loss
    assert best_validation_loss == best_val_loss
    assert trainer.best_validation_metric == best_val_loss, (
        "Best validation loss was not correctly restored"
    )

    # Validate that state is properly loaded -- restart from the following epoch
    assert trainer.current_epoch == epoch + 1, "Epoch should be restored from checkpoint"

    # Validate model parameters
    assert equal_models(trainer.model, loaded_model), "Model checkpoint not loaded correctly"

    # Validate optimizer state
    assert trainer.optimizer is not None, "Optimizer should be initialized"
    assert_equal_optimizer_states(loaded_optim.state_dict(), trainer.optimizer.state_dict())

    # Validate LR scheduler state
    assert trainer.lr_scheduler.state_dict() == loaded_lr_scheduler.state_dict(), (
        "LR scheduler state mismatch"
    )

    # Validate RNG state
    assert trainer.random_seed == random_seed, "Random seed was not restored"
    assert torch.equal(torch_rng_state, trainer.torch_rng_state), "Torch RNG state mismatch"


if __name__ == "__main__":
    import torch
    from torch.nn import Linear

    model1 = Linear(in_features=10, out_features=1, bias=True)
    model2 = Linear(in_features=10, out_features=1, bias=True)
    model2.load_state_dict(model1.state_dict())
    model1 = model1.to("cuda")
    print(equal_models(model1, model2))



import os
from pathlib import Path
from typing import Generator
from unittest.mock import patch

import pytest

from itwinai.slurm.slurm_script_builder import (
    SlurmScriptBuilder,
    SlurmScriptConfiguration,
)


@pytest.fixture
def cd_tmpdir(tmp_path) -> Generator[None, None, None]:
    old_cwd = Path.cwd()
    os.chdir(tmp_path)
    yield
    os.chdir(old_cwd)


@pytest.fixture
def slurm_config():
    return SlurmScriptConfiguration(
        job_name="test_job",
        account="test_account",
        partition="test_partition",
        time="01:00:00",
        std_out=Path("slurm_job_logs/test_job.out"),
        err_out=Path("slurm_job_logs/test_job.err"),
        num_nodes=2,
        num_tasks_per_node=4,
        gpus_per_node=2,
        cpus_per_gpu=8,
    )


@pytest.fixture
def slurm_builder(slurm_config: SlurmScriptConfiguration):
    return SlurmScriptBuilder(
        slurm_script_configuration=slurm_config,
        distributed_strategy="ddp",
        training_command="python train.py",
    )


@pytest.mark.parametrize(
    "save_script, submit_slurm_job",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_process_slurm_script(
    slurm_builder: SlurmScriptBuilder,
    save_script: bool,
    submit_slurm_job: bool,
    tmp_path: Path,
    cd_tmpdir: None,
):
    """Test that process_slurm_script behaves correctly for all cases of save_script
    and submit_slurm_job."""

    file_path = tmp_path / "slurm_script.sh"

    with patch("subprocess.run") as mock_run:
        slurm_builder.process_slurm_script(
            file_path=file_path,
            save_script=save_script,
            submit_slurm_job=submit_slurm_job,
        )

        # Checking that it creates the stdout and stderr directories
        assert slurm_builder.slurm_script_configuration.std_out is not None
        assert slurm_builder.slurm_script_configuration.std_out.parent.exists()
        assert slurm_builder.slurm_script_configuration.err_out is not None
        assert slurm_builder.slurm_script_configuration.err_out.parent.exists()

        # Check if sbatch was called
        if submit_slurm_job:
            mock_run.assert_called_once()
            called_args = mock_run.call_args[0][0]
            assert called_args[0] == "sbatch"
        else:
            mock_run.assert_not_called()

        # Check if the script file exists
        assert file_path.exists() == save_script


def test_process_slurm_script_twice(
    slurm_builder: SlurmScriptBuilder, tmp_path: Path, cd_tmpdir: None
):
    """Ensure that calling process_slurm_script twice with save_script=True fails,
    but calling it first with save_script=True and then with save_script=False works fine.
    """

    file_path = Path(tmp_path) / "slurm_script.sh"

    # First call with save_script=True should succeed
    slurm_builder.process_slurm_script(
        file_path=file_path, save_script=True, submit_slurm_job=False
    )
    assert file_path.exists()  # The script should be saved

    # Second call with save_script=False should not fail
    slurm_builder.process_slurm_script(
        file_path=file_path, save_script=False, submit_slurm_job=False
    )
    assert file_path.exists()  # The script should still be there

    # Second call with save_script=True should fail due to file already existing
    with pytest.raises(FileExistsError):
        slurm_builder.process_slurm_script(
            file_path=file_path, save_script=True, submit_slurm_job=False
        )
